import logging

import fontTools.designspaceLib as designspaceLib
import pytest
from fontTools.pens.recordingPen import RecordingPen

import ufo2ft.instantiator
from ufo2ft.util import openFont, openFontFactory


def test_interpolation_weight_width_class(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    for instance in designspace.instances:
        instance.font = generator.generate_instance(instance)

    # LightCondensed
    font = designspace.instances[0].font
    assert font.info.openTypeOS2WeightClass == 1
    assert font.info.openTypeOS2WidthClass == 1

    # BoldCondensed
    font = designspace.instances[1].font
    assert font.info.openTypeOS2WeightClass == 1000
    assert font.info.openTypeOS2WidthClass == 1

    # LightWide
    font = designspace.instances[2].font
    assert font.info.openTypeOS2WeightClass == 1
    assert font.info.openTypeOS2WidthClass == 9

    # BoldWide
    font = designspace.instances[3].font
    assert font.info.openTypeOS2WeightClass == 1000
    assert font.info.openTypeOS2WidthClass == 9

    # Medium_Narrow_I
    font = designspace.instances[4].font
    assert font.info.openTypeOS2WeightClass == 500
    assert font.info.openTypeOS2WidthClass == 9

    # Medium_Wide_I
    font = designspace.instances[5].font
    assert font.info.openTypeOS2WeightClass == 500
    assert font.info.openTypeOS2WidthClass == 9

    # Two
    font = designspace.instances[6].font
    assert font.info.openTypeOS2WeightClass == 1000
    assert font.info.openTypeOS2WidthClass == 9

    # One
    font = designspace.instances[7].font
    assert font.info.openTypeOS2WeightClass == 500
    assert font.info.openTypeOS2WidthClass == 9


def test_default_groups_only(ufo_module, data_dir, caplog):
    """Test that only the default source's groups end up in instances."""

    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )
    d.addSourceDescriptor(location={"Weight": 300}, font=ufo_module.Font())
    d.addSourceDescriptor(location={"Weight": 900}, font=ufo_module.Font())
    d.addInstanceDescriptor(styleName="2", location={"Weight": 400})

    d.sources[0].font.groups["public.kern1.GRK_alpha_alt_LC_1ST"] = [
        "alpha.alt",
        "alphatonos.alt",
    ]
    d.sources[1].font.groups["public.kern1.GRK_alpha_LC_1ST"] = [
        "alpha.alt",
        "alphatonos.alt",
    ]

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    assert "contains different groups than the default source" in caplog.text

    instance = generator.generate_instance(d.instances[0])
    assert instance.groups == {
        "public.kern1.GRK_alpha_alt_LC_1ST": ["alpha.alt", "alphatonos.alt"]
    }


def test_default_groups_only2(ufo_module, data_dir, caplog):
    """Test that the group difference warning is not triggered if non-default
    source groups are empty."""

    d = designspaceLib.DesignSpaceDocument()
    d.addAxisDescriptor(
        name="Weight", tag="wght", minimum=300, default=300, maximum=900
    )
    d.addSourceDescriptor(location={"Weight": 300}, font=ufo_module.Font())
    d.addSourceDescriptor(location={"Weight": 900}, font=ufo_module.Font())
    d.addInstanceDescriptor(styleName="2", location={"Weight": 400})

    d.sources[0].font.groups["public.kern1.GRK_alpha_alt_LC_1ST"] = [
        "alpha.alt",
        "alphatonos.alt",
    ]

    generator = ufo2ft.instantiator.Instantiator.from_designspace(d)
    assert "contains different groups than the default source" not in caplog.text

    instance = generator.generate_instance(d.instances[0])
    assert instance.groups == {
        "public.kern1.GRK_alpha_alt_LC_1ST": ["alpha.alt", "alphatonos.alt"]
    }


def test_interpolation_no_rounding(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.instances[4].location = {"weight": 123.456, "width": 789.123}
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=False
    )

    instance_font = generator.generate_instance(designspace.instances[4])
    assert isinstance(instance_font.info.ascender, float)
    assert isinstance(instance_font.kerning[("A", "J")], float)
    assert isinstance(instance_font["A"].contours[0][0].x, float)


def test_interpolation_rounding(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.instances[4].location = {"weight": 123.456, "width": 789.123}
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[4])
    assert isinstance(instance_font.info.ascender, int)
    assert isinstance(instance_font.kerning[("A", "J")], int)
    assert isinstance(instance_font["A"].contours[0][0].x, int)


def test_weight_class_from_wght_axis():
    assert ufo2ft.instantiator.weight_class_from_wght_value(-500) == 1
    assert ufo2ft.instantiator.weight_class_from_wght_value(1.1) == 1
    assert ufo2ft.instantiator.weight_class_from_wght_value(1) == 1
    assert ufo2ft.instantiator.weight_class_from_wght_value(500.6) == 501
    assert ufo2ft.instantiator.weight_class_from_wght_value(1000) == 1000
    assert ufo2ft.instantiator.weight_class_from_wght_value(1000.0) == 1000
    assert ufo2ft.instantiator.weight_class_from_wght_value(1000.1) == 1000
    assert ufo2ft.instantiator.weight_class_from_wght_value(2000.1) == 1000


def test_width_class_from_wdth_axis():
    assert ufo2ft.instantiator.width_class_from_wdth_value(-500) == 1
    assert ufo2ft.instantiator.width_class_from_wdth_value(50) == 1
    assert ufo2ft.instantiator.width_class_from_wdth_value(62.5) == 2
    assert ufo2ft.instantiator.width_class_from_wdth_value(75) == 3
    assert ufo2ft.instantiator.width_class_from_wdth_value(87.5) == 4
    assert ufo2ft.instantiator.width_class_from_wdth_value(100) == 5
    assert ufo2ft.instantiator.width_class_from_wdth_value(112) == 6
    assert ufo2ft.instantiator.width_class_from_wdth_value(112.5) == 6
    assert ufo2ft.instantiator.width_class_from_wdth_value(125) == 7
    assert ufo2ft.instantiator.width_class_from_wdth_value(130) == 7
    assert ufo2ft.instantiator.width_class_from_wdth_value(150) == 8
    assert ufo2ft.instantiator.width_class_from_wdth_value(190) == 9
    assert ufo2ft.instantiator.width_class_from_wdth_value(200) == 9
    assert ufo2ft.instantiator.width_class_from_wdth_value(1000) == 9


def test_swap_glyph_names(ufo_module, data_dir):
    ufo = openFont(data_dir / "SwapGlyphNames" / "A.ufo", ufo_module=ufo_module)

    ufo2ft.instantiator.swap_glyph_names(ufo, "a", "a.swap")

    # Test swapped outlines.
    assert ufo["a"].unicode == 0x61
    assert len(ufo["a"]) == 1
    assert len(ufo["a"][0]) == 8
    assert ufo["a"].width == 666
    assert ufo["a.swap"].unicode is None
    assert len(ufo["a.swap"]) == 1
    assert len(ufo["a.swap"][0]) == 4
    assert ufo["a.swap"].width == 600

    # Test swapped components.
    assert sorted(c.baseGlyph for c in ufo["aaa"].components) == [
        "a.swap",
        "a.swap",
        "x",
    ]
    assert sorted(c.baseGlyph for c in ufo["aaa.swap"].components) == ["a", "a", "y"]

    # Test swapped anchors.
    assert [dict(a) for a in ufo["a"].anchors] == [
        dict(x=153, y=0, name="bottom"),
        dict(x=153, y=316, name="top"),
    ]
    assert [dict(a) for a in ufo["a.swap"].anchors] == [
        dict(x=351, y=0, name="bottom"),
        dict(x=351, y=613, name="top"),
    ]

    # Test swapped glyph kerning.
    assert ufo.kerning == {
        ("public.kern1.a", "x"): 10,
        ("public.kern1.aswap", "x"): 20,
        ("a", "y"): 40,
        ("a.swap", "y"): 30,
        ("y", "a"): 60,
        ("y", "a.swap"): 50,
    }

    # Test swapped group membership.
    assert ufo.groups == {
        "public.kern1.a": ["a.swap"],
        "public.kern1.aswap": ["a"],
        "public.kern2.a": ["a.swap", "a"],
    }

    # Swap a second time.
    ufo2ft.instantiator.swap_glyph_names(ufo, "aaa", "aaa.swap")

    # Test swapped glyphs.
    assert sorted(c.baseGlyph for c in ufo["aaa"].components) == ["a", "a", "y"]
    assert sorted(c.baseGlyph for c in ufo["aaa.swap"].components) == [
        "a.swap",
        "a.swap",
        "x",
    ]

    # Test for no leftover temporary glyphs.
    assert {g.name for g in ufo} == {
        "space",
        "a",
        "a.swap",
        "aaa",
        "aaa.swap",
        "x",
        "y",
    }

    with pytest.raises(ufo2ft.instantiator.InstantiatorError, match="Cannot swap"):
        ufo2ft.instantiator.swap_glyph_names(ufo, "aaa", "aaa.swapa")


def test_swap_glyph_names_spec(ufo_module, data_dir):
    """Test that the rule example in the designspaceLib spec works.

    `adieresis` should look the same as before the rule application.

    [1]: fonttools/Doc/source/designspaceLib#ufo-instances
    """
    ufo = openFont(data_dir / "SwapGlyphNames" / "B.ufo", ufo_module=ufo_module)
    ufo2ft.instantiator.swap_glyph_names(ufo, "a", "a.alt")

    assert sorted(c.baseGlyph for c in ufo["adieresis"].components) == [
        "a.alt",
        "dieresiscomb",
    ]
    assert sorted(c.baseGlyph for c in ufo["adieresis.alt"].components) == [
        "a",
        "dieresiscomb",
    ]


def test_rules_are_applied_deterministically(ufo_module, data_dir):
    """Test that a combination of designspace rules that end up mapping
    serveral input glyphs to the same destination glyph result in a correct and
    deterministic series of glyph swaps.

    The example is a font with 2 Q designs that depend on a style axis
        style < 0.5: Q        style >= 0.5: Q.ss01
    and each Q also has an alternative shape in bolder weights (like Skia)
        weight < 780: Q       weight >= 780: Q.alt
        weight < 730: Q.ss01  weight >= 730: Q.ss01.alt

    Then we generate an instance at style = 1, weight = 900. From the rules,
    the default CMAP entry for Q should have the outlines of Q.ss01.alt from
    the black UFO.
    """
    doc = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceRuleOrder" / "MyFont.designspace"
    )
    doc.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(doc)
    instance = generator.generate_instance(doc.instances[0])
    pen = RecordingPen()
    instance["Q"].draw(pen)
    instance_recording = pen.value

    black_ufo = openFont(
        data_dir / "DesignspaceRuleOrder" / "MyFont_Black.ufo", ufo_module=ufo_module
    )
    pen = RecordingPen()
    black_ufo["Q.ss01.alt"].draw(pen)
    black_ufo_recording = pen.value

    assert instance_recording == black_ufo_recording


def test_raise_no_default_master(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans_no_default.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))

    with pytest.raises(ufo2ft.instantiator.InstantiatorError, match="no default"):
        ufo2ft.instantiator.Instantiator.from_designspace(
            designspace, round_geometry=True
        )


def test_raise_failed_glyph_interpolation(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceBrokenTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(designspace)

    with pytest.raises(
        ufo2ft.instantiator.InstantiatorError, match="Failed to generate instance"
    ):
        for instance in designspace.instances:
            instance.font = generator.generate_instance(instance)


def test_ignore_failed_glyph_interpolation(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceBrokenTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(designspace)
    generator.skip_export_glyphs.append("asas")

    for instance in designspace.instances:
        instance.font = generator.generate_instance(instance)
        assert (
            not instance.font["asas"].contours and not instance.font["asas"].components
        )


def test_raise_anisotropic_location(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans-width-only.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.instances[0].location["width"] = (100, 900)

    with pytest.raises(
        ufo2ft.instantiator.InstantiatorError, match="anisotropic instance locations"
    ):
        ufo2ft.instantiator.Instantiator.from_designspace(
            designspace, round_geometry=True
        )


def test_copy_nonkerning_group(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(designspace)

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.groups == {
        "nonkerning_group": ["A"],
        "public.kern2.asdf": ["A"],
    }


def test_interpolation(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font["l"].width == 220


def test_interpolation_only_default(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    for name in designspace.default.font.glyphOrder:
        if name != "A":
            del designspace.default.font[name]

    with caplog.at_level(logging.WARNING):
        generator = ufo2ft.instantiator.Instantiator.from_designspace(
            designspace, round_geometry=True
        )
    assert "contains glyphs that are missing from the" in caplog.text

    instance_font = generator.generate_instance(designspace.instances[0])
    assert {g.name for g in instance_font} == {"A"}


def test_interpolation_masters_as_instances(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir
        / "DesignspaceBrokenTest"
        / "Designspace-MastersAsInstances.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.styleName == "Light ASDF"
    assert instance_font["l"].width == 160
    instance_font = generator.generate_instance(designspace.instances[1])
    assert instance_font.info.styleName == "Bold ASDF"
    assert instance_font["l"].width == 280


def test_non_default_layer(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSans" / "MutatorSans-non-default-layer.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    instance_font = generator.generate_instance(designspace.instances[0])
    assert {g.name for g in instance_font} == {"A", "S", "W"}


def test_instance_attributes(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-instance-attrs.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.familyName == "aaa"
    assert instance_font.info.styleName == "sss"
    assert instance_font.info.postscriptFontName == "ppp"
    assert instance_font.info.styleMapFamilyName == "yyy"
    assert instance_font.info.styleMapStyleName == "xxx"


def test_instance_no_attributes(ufo_module, data_dir, caplog):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-bare.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    with caplog.at_level(logging.WARNING):
        instance_font = generator.generate_instance(designspace.instances[0])
    assert "missing the stylename attribute" in caplog.text

    assert instance_font.info.familyName == "MyFont"
    assert instance_font.info.styleName == "Light"
    assert instance_font.info.postscriptFontName is None
    assert instance_font.info.styleMapFamilyName is None
    assert instance_font.info.styleMapStyleName is None


def test_axis_mapping(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-wght-wdth.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass == 400
    assert instance_font.info.openTypeOS2WidthClass == 5
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [
        ("weight", 100.0),
        ("width", 100.0),
    ]


def test_axis_mapping_manual_os2_classes(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-wght-wdth.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.sources[0].font.info.openTypeOS2WeightClass = 800
    designspace.sources[0].font.info.openTypeOS2WidthClass = 7
    designspace.sources[1].font.info.openTypeOS2WeightClass = 900
    designspace.sources[1].font.info.openTypeOS2WidthClass = 9
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass == 850
    assert instance_font.info.openTypeOS2WidthClass == 8
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [
        ("weight", 100.0),
        ("width", 100.0),
    ]


def test_axis_mapping_no_os2_width_class_inference(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-bare.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    designspace.sources[0].font.info.openTypeOS2WeightClass = 800
    designspace.sources[1].font.info.openTypeOS2WeightClass = 900
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass == 850
    assert instance_font.info.openTypeOS2WidthClass is None
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [("weight", 100.0)]


def test_axis_mapping_no_os2_class_inference(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-opsz.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass is None
    assert instance_font.info.openTypeOS2WidthClass is None
    assert instance_font.info.italicAngle is None
    assert instance_font.lib["designspace.location"] == [("optical", 15.0)]


def test_axis_mapping_italicAngle_inference(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-slnt.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.info.openTypeOS2WeightClass is None
    assert instance_font.info.openTypeOS2WidthClass is None
    assert instance_font.info.italicAngle == 40.123
    assert instance_font.lib["designspace.location"] == [("slant", 40.123)]


def test_lib_into_instance(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest-lib.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )

    assert designspace.default.font.lib["blorb"] == "asasa"
    assert "public.skipExportGlyphs" not in designspace.sources[0].font.lib

    instance_font = generator.generate_instance(designspace.instances[0])
    assert instance_font.lib["blorb"] == "asasa"
    assert instance_font.lib["public.skipExportGlyphs"] == ["a", "b", "c"]

    instance_font2 = generator.generate_instance(designspace.instances[1])
    assert instance_font2.lib["blorb"] == "asasa"
    assert instance_font2.lib["public.skipExportGlyphs"] == ["a", "b", "c"]


def test_data_independence(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "DesignspaceTest" / "DesignspaceTest.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    instance_font1 = generator.generate_instance(designspace.instances[0])
    designspace.instances[0].lib["aaaaaaaa"] = 1
    instance_font2 = generator.generate_instance(designspace.instances[0])

    instance_font1["l"].unicodes.append(2)
    assert instance_font1["l"].unicodes == [0x6C, 2]
    assert instance_font2["l"].unicodes == [0x6C]

    instance_font1["l"].lib["asdf"] = 1
    assert instance_font1["l"].lib == {"asdf": 1}
    assert not instance_font2["l"].lib

    generator.copy_lib["sdjkhsjdhjdf"] = 1
    instance_font1.lib["asdf"] = 1
    assert instance_font1.lib == {
        "asdf": 1,
        "blorb": "asasa",
        "designspace.location": [("weight", 100.0)],
        "public.skipExportGlyphs": [],
    }
    assert instance_font2.lib == {
        "blorb": "asasa",
        "designspace.location": [("weight", 100.0)],
        "public.skipExportGlyphs": [],
    }

    assert generator.copy_info.openTypeOS2Panose == [2, 11, 5, 4, 2, 2, 2, 2, 2, 4]
    generator.copy_info.openTypeOS2Panose.append(1000)
    assert instance_font1.info.openTypeOS2Panose is None
    assert instance_font2.info.openTypeOS2Panose is None

    # copy_feature_text not tested because it is a(n immutable) string

    assert not generator.skip_export_glyphs
    generator.skip_export_glyphs.extend(["a", "b"])
    assert not instance_font1.lib["public.skipExportGlyphs"]
    assert not instance_font2.lib["public.skipExportGlyphs"]
    instance_font1.lib["public.skipExportGlyphs"].append("z")
    assert not instance_font2.lib["public.skipExportGlyphs"]


def test_skipped_fontinfo_attributes():
    """Test that we consider all available font info attributes for copying."""
    import fontMath.mathInfo
    import fontTools.ufoLib

    SKIPPED_ATTRS = {
        "guidelines",
        "macintoshFONDFamilyID",
        "macintoshFONDName",
        "openTypeNameCompatibleFullName",
        "openTypeNamePreferredFamilyName",
        "openTypeNamePreferredSubfamilyName",
        "openTypeNameUniqueID",
        "openTypeNameWWSFamilyName",
        "openTypeNameWWSSubfamilyName",
        "openTypeOS2Panose",
        "postscriptFontName",
        "postscriptFullName",
        "postscriptUniqueID",
        "styleMapFamilyName",
        "styleMapStyleName",
        "styleName",
        "woffMetadataUniqueID",
        "year",
    }

    assert (
        fontTools.ufoLib.fontInfoAttributesVersion3
        - set(fontMath.mathInfo._infoAttrs.keys())
        - {"postscriptWeightName"}  # Handled in fontMath specially.
        - ufo2ft.instantiator.UFO_INFO_ATTRIBUTES_TO_COPY_TO_INSTANCES
        == SKIPPED_ATTRS
    )


def test_designspace_v5_discrete_axis_raises_error(data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "MutatorSansLite" / "MutatorFamily_v5_discrete_axis.designspace"
    )
    # The error message should advise to use `splitInterpolable()`
    with pytest.raises(
        ufo2ft.instantiator.InstantiatorError, match="splitInterpolable"
    ):
        ufo2ft.instantiator.Instantiator.from_designspace(designspace)


def test_strict_math_glyph(ufo_module, data_dir):
    designspace = designspaceLib.DesignSpaceDocument.fromfile(
        data_dir / "InstantiatorStrictMathGlyph" / "StrictMathGlyph.designspace"
    )
    designspace.loadSourceFonts(openFontFactory(ufo_module=ufo_module))
    generator = ufo2ft.instantiator.Instantiator.from_designspace(
        designspace, round_geometry=True
    )
    fonts = [
        generator.generate_instance(instance) for instance in designspace.instances
    ]
    assert len(fonts) == 1
    glyph = fonts[0]["test"]
    assert len(glyph.contours) == 1
    assert len(glyph.contours[0].points) == 16



from ufo2ft.featureWriters import (
    FEATURE_WRITERS_KEY,
    BaseFeatureWriter,
    loadFeatureWriterFromString,
    loadFeatureWriters,
)

try:
    from plistlib import FMT_XML, loads

    def readPlistFromString(s):
        return loads(s, fmt=FMT_XML)

except ImportError:
    from plistlib import readPlistFromString

import pytest

from ..testSupport import _TempModule

TEST_LIB_PLIST = readPlistFromString(
    b"""
<dict>
    <key>com.github.googlei18n.ufo2ft.featureWriters</key>
    <array>
        <dict>
            <key>class</key>
            <string>KernFeatureWriter</string>
            <key>options</key>
            <dict>
                <key>mode</key>
                <string>skip</string>
            </dict>
        </dict>
    </array>
</dict>
"""
)


class FooBarWriter(BaseFeatureWriter):
    tableTag = "GSUB"

    def __init__(self, **kwargs):
        pass

    def write(self, font, feaFile, compiler=None):
        return False


@pytest.fixture(scope="module", autouse=True)
def customWriterModule():
    """Make a temporary 'myFeatureWriters' module containing a
    'FooBarWriter' class for testing the wruter loading machinery.
    """
    with _TempModule("myFeatureWriters") as temp_module:
        temp_module.module.__dict__["FooBarWriter"] = FooBarWriter
        yield


VALID_SPEC_LISTS = [
    [{"class": "KernFeatureWriter"}],
    [
        {"class": "KernFeatureWriter", "options": {"ignoreMarks": False}},
        {"class": "MarkFeatureWriter", "options": {"features": ["mark"]}},
    ],
    [{"class": "FooBarWriter", "module": "myFeatureWriters", "options": {"a": 1}}],
    TEST_LIB_PLIST[FEATURE_WRITERS_KEY],
]


@pytest.mark.parametrize("specList", VALID_SPEC_LISTS)
def test_loadFeatureWriters_valid(specList, FontClass):
    ufo = FontClass()
    ufo.lib[FEATURE_WRITERS_KEY] = specList
    for writer in loadFeatureWriters(ufo, ignoreErrors=False):
        assert writer.tableTag in {"GSUB", "GPOS"}
        assert callable(writer.write)


VALID_SPEC_STRINGS = [
    "KernFeatureWriter",
    "KernFeatureWriter(ignoreMarks=False)",
    "MarkFeatureWriter(features=['mark'])",
    "myFeatureWriters::FooBarWriter(a=1)",
]


@pytest.mark.parametrize("spec", VALID_SPEC_STRINGS)
def test_loadFeatureWriterFromString_valid(spec, FontClass):
    writer = loadFeatureWriterFromString(spec)
    assert writer.tableTag in {"GSUB", "GPOS"}
    assert callable(writer.write)



"""This is a redirection stub, because the original module was
misnamed."""

import warnings

from .dottedCircle import DottedCircleFilter  # noqa

warnings.warn(
    "The dottedCircleFilter module is deprecated, please import dottedCircle instead.",
    UserWarning,
    stacklevel=1,
)



from __future__ import annotations

import logging
import os
import re
from collections import OrderedDict
from inspect import isclass
from io import StringIO
from tempfile import NamedTemporaryFile

from fontTools import mtiLib
from fontTools.designspaceLib import DesignSpaceDocument, SourceDescriptor
from fontTools.feaLib.builder import addOpenTypeFeaturesFromString
from fontTools.feaLib.error import FeatureLibError, IncludedFeaNotFound
from fontTools.feaLib.parser import Parser
from fontTools.misc.loggingTools import Timer

from ufo2ft.constants import MTI_FEATURES_PREFIX
from ufo2ft.featureWriters import (
    CursFeatureWriter,
    GdefFeatureWriter,
    KernFeatureWriter,
    MarkFeatureWriter,
    ast,
    isValidFeatureWriter,
    loadFeatureWriters,
)
from ufo2ft.util import describe_ufo

logger = logging.getLogger(__name__)
timer = Timer(logging.getLogger("ufo2ft.timer"), level=logging.DEBUG)


def parseLayoutFeatures(font, includeDir=None):
    """Parse OpenType layout features in the UFO and return a
    feaLib.ast.FeatureFile instance.

    includeDir is an optional directory path to search for included
    feature files, if omitted the font.path is used. If the latter
    is also not set, the feaLib Lexer uses the current working directory.
    """
    featxt = font.features.text or ""
    if not featxt:
        return ast.FeatureFile()
    buf = StringIO(featxt)
    ufoPath = font.path
    if includeDir is None and ufoPath is not None:
        # The UFO v3 specification says "Any include() statements must be relative to
        # the UFO path, not to the features.fea file itself". We set the `name`
        # attribute on the buffer to the actual feature file path, which feaLib will
        # pick up and use to attribute errors to the correct file, and explicitly set
        # the include directory to the parent of the UFO.
        ufoPath = os.path.normpath(ufoPath)
        buf.name = os.path.join(ufoPath, "features.fea")
        includeDir = os.path.dirname(ufoPath) or "."
    glyphNames = set(font.keys())
    includeDir = os.path.normpath(includeDir) if includeDir else None
    try:
        parser = Parser(buf, glyphNames, includeDir=includeDir)
        doc = parser.parse()
    except IncludedFeaNotFound as e:
        if ufoPath and os.path.exists(os.path.join(ufoPath, e.args[0])):
            logger.warning(
                "Please change the file name in the include(...); "
                "statement to be relative to the UFO itself, "
                "instead of relative to the 'features.fea' file "
                "contained in it."
            )
        raise
    return doc


class BaseFeatureCompiler:
    """Base class for generating OpenType features and compiling OpenType
    layout tables from these.
    """

    def __init__(self, ufo, ttFont=None, glyphSet=None, extraSubstitutions=None):
        """
        Args:
          ufo: an object representing a UFO (defcon.Font or equivalent)
            containing the features source data.
          ttFont: a fontTools TTFont object where the generated OpenType
            tables are added. If None, an empty TTFont is used, with
            the same glyph order as the ufo object.
          glyphSet: a (optional) dict containing pre-processed copies of
            the UFO glyphs.
          extraSubstitutions: an optional dictionary mapping glyph names
            to a set of other glyphs which should be considered reachable
            from them (for example when using designspace rules to effect
            substitutions).
        """
        self.ufo = ufo

        if ttFont is None:
            from fontTools.ttLib import TTFont

            from ufo2ft.util import makeOfficialGlyphOrder

            ttFont = TTFont()
            ttFont.setGlyphOrder(makeOfficialGlyphOrder(ufo))
        self.ttFont = ttFont

        glyphOrder = ttFont.getGlyphOrder()
        if glyphSet is not None:
            if set(glyphOrder) != set(glyphSet.keys()):
                print("Glyph order incompatible")
                print("In UFO but not in font:", set(glyphSet.keys()) - set(glyphOrder))
                print("In font but not in UFO:", set(glyphOrder) - set(glyphSet.keys()))
            assert set(glyphOrder) == set(glyphSet.keys())
        else:
            glyphSet = ufo
        self.glyphSet = OrderedDict((gn, glyphSet[gn]) for gn in glyphOrder)

        self.extraSubstitutions = extraSubstitutions

    def setupFeatures(self):
        """Make the features source.

        **This should not be called externally.** Subclasses
        must override this method.
        """
        raise NotImplementedError

    def buildTables(self):
        """Compile OpenType feature tables from the source.

        **This should not be called externally.** Subclasses
        must override this method.
        """
        raise NotImplementedError

    def setupFile_features(self):
        """DEPRECATED. Use 'setupFeatures' instead."""
        _deprecateMethod("setupFile_features", "setupFeatures")
        self.setupFeatures()

    def setupFile_featureTables(self):
        """DEPRECATED. Use 'setupFeatures' instead."""
        _deprecateMethod("setupFile_featureTables", "buildTables")
        self.buildTables()

    def compile(self):
        if "setupFile_features" in self.__class__.__dict__:
            _deprecateMethod("setupFile_features", "setupFeatures")
            self.setupFile_features()
        else:
            self.setupFeatures()

        if "setupFile_featureTables" in self.__class__.__dict__:
            _deprecateMethod("setupFile_featureTables", "buildTables")
            self.setupFile_featureTables()
        else:
            self.buildTables()

        return self.ttFont


def _deprecateMethod(arg, repl):
    import warnings

    warnings.warn(
        f"{arg!r} method is deprecated; use {repl!r} instead",
        category=UserWarning,
        stacklevel=3,
    )


class FeatureCompiler(BaseFeatureCompiler):
    """Generate automatic features and compile OpenType tables from Adobe
    Feature File stored in the UFO, using fontTools.feaLib as compiler.
    """

    defaultFeatureWriters = [
        CursFeatureWriter,
        KernFeatureWriter,
        MarkFeatureWriter,
        GdefFeatureWriter,
    ]

    def __init__(
        self,
        ufo,
        ttFont=None,
        glyphSet=None,
        featureWriters=None,
        feaIncludeDir=None,
        extraSubstitutions=None,
        **kwargs,
    ):
        """
        Args:
          featureWriters: a list of BaseFeatureWriter subclasses or
            pre-initialized instances. The default value (None) means that:
            - first, the UFO lib will be searched for a list of featureWriters
              under the key "com.github.googlei18n.ufo2ft.featureWriters"
              (see loadFeatureWriters).
            - if that is not found, the default list of writers will be used:
              (see FeatureCompiler.defaultFeatureWriters, and the individual
              feature writer classes for the list of features generated).
            If the featureWriters list is empty, no automatic feature is
            generated and only pre-existing features are compiled.
            The ``featureWriters`` parameter overrides both the writers from
            the UFO lib and the default writers list. To extend instead of
            replace the latter, the list can contain a special value ``...``
            (i.e. the ``ellipsis`` singleton, not the str literal '...')
            which gets replaced by either the UFO.lib writers or the default
            ones; thus one can insert additional writers either before or after
            these.
          feaIncludeDir: a directory to be used as the include directory for
            the feature file. If None, the include directory is set to the
            parent directory of the UFO, provided the UFO has a path.
        """
        BaseFeatureCompiler.__init__(
            self, ufo, ttFont, glyphSet, extraSubstitutions=extraSubstitutions
        )
        self.feaIncludeDir = feaIncludeDir

        self.initFeatureWriters(featureWriters)

        if kwargs.get("mtiFeatures") is not None:
            import warnings

            warnings.warn(
                "mtiFeatures argument is ignored; "
                "you should use MtiLibFeatureCompiler",
                category=UserWarning,
                stacklevel=2,
            )

    def _load_custom_feature_writers(self, featureWriters=None):
        # Args:
        #   ufo: Font
        #   featureWriters: Optional[List[Union[FeatureWriter, EllipsisType]]])
        # Returns: List[FeatureWriter]

        # by default, load the feature writers from the lib or the default ones;
        # ellipsis is used as a placeholder so one can optionally insert additional
        # featureWriters=[w1, ..., w2] either before or after these, or override
        # them by omitting the ellipsis.
        if featureWriters is None:
            featureWriters = [...]
        result = []
        seen_ellipsis = False
        for writer in featureWriters:
            if writer is ...:
                if seen_ellipsis:
                    raise ValueError("ellipsis not allowed more than once")
                writers = loadFeatureWriters(self.ufo)
                if writers is not None:
                    result.extend(writers)
                else:
                    result.extend(self.defaultFeatureWriters)
                seen_ellipsis = True
            else:
                klass = writer if isclass(writer) else type(writer)
                if not isValidFeatureWriter(klass):
                    raise TypeError(f"Invalid feature writer: {writer!r}")
                result.append(writer)
        return result

    def initFeatureWriters(self, featureWriters=None):
        """Initialize feature writer classes as specified in the UFO lib.
        If none are defined in the UFO, the default feature writers are used
        (see FeatureCompiler.defaultFeatureWriters).
        The 'featureWriters' argument can be used to override these.
        The method sets the `self.featureWriters` attribute with the list of
        writers.

        Note that the writers that generate GSUB features are placed first in
        this list, before all others. This is because the GSUB table may be
        used in the subsequent feature writers to resolve substitutions from
        glyphs with unicodes to their alternates.
        """
        featureWriters = self._load_custom_feature_writers(featureWriters)

        gsubWriters = []
        others = []
        for writer in featureWriters:
            if isclass(writer):
                writer = writer()
            if writer.tableTag == "GSUB":
                gsubWriters.append(writer)
            else:
                others.append(writer)

        self.featureWriters = gsubWriters + others

    def setupFeatures(self):
        """
        Make the features source.

        **This should not be called externally.** Subclasses
        may override this method to handle the file creation
        in a different way if desired.
        """
        with timer("run feature writers"):
            if self.featureWriters:
                featureFile = parseLayoutFeatures(self.ufo, self.feaIncludeDir)

                # Insertion markers are only considered in "skip" mode.
                if any(writer.mode == "skip" for writer in self.featureWriters):
                    markers = {
                        writer.insertFeatureMarker
                        for writer in self.featureWriters
                        if writer.insertFeatureMarker is not None
                    }
                    warn_about_miscased_insertion_markers(
                        describe_ufo(self.ufo), featureFile, markers
                    )

                path = self.ufo.path
                for writer in self.featureWriters:
                    try:
                        writer.write(self.ufo, featureFile, compiler=self)
                    except FeatureLibError:
                        if path is None:
                            self._write_temporary_feature_file(featureFile.asFea())
                        raise

                # stringify AST to get correct line numbers in error messages
                self.features = featureFile.asFea()
            else:
                # no featureWriters, simply read existing features' text
                self.features = self.ufo.features.text or ""

    def writeFeatures(self, outfile):
        if hasattr(self, "features"):
            outfile.write(self.features)

    def buildTables(self):
        """
        Compile OpenType feature tables from the source.
        Raises a FeaLibError if the feature compilation was unsuccessful.

        **This should not be called externally.** Subclasses
        may override this method to handle the table compilation
        in a different way if desired.
        """

        if not self.features:
            return

        # the path is used by the lexer to follow 'include' statements;
        # if we generated some automatic features, includes have already been
        # resolved, and we work from a string which does't exist on disk
        path = self.ufo.path if not self.featureWriters else None
        with timer("build OpenType features"):
            try:
                addOpenTypeFeaturesFromString(self.ttFont, self.features, filename=path)
            except FeatureLibError:
                if path is None:
                    self._write_temporary_feature_file(self.features)
                raise

    def _write_temporary_feature_file(self, features: str) -> None:
        # if compilation fails, create temporary file for inspection
        data = features.encode("utf-8")
        with NamedTemporaryFile(delete=False) as tmp:
            tmp.write(data)
        logger.error("Compilation failed! Inspect temporary file: %r", tmp.name)


class MtiFeatureCompiler(BaseFeatureCompiler):
    """Compile OpenType layout tables from MTI feature files using
    fontTools.mtiLib.
    """

    def setupFeatures(self):
        ufo = self.ufo
        features = {}
        # includes the length of the "/" separator
        prefixLength = len(MTI_FEATURES_PREFIX) + 1
        for fn in ufo.data.fileNames:
            if fn.startswith(MTI_FEATURES_PREFIX) and fn.endswith(".mti"):
                content = ufo.data[fn].decode("utf-8")
                features[fn[prefixLength:-4]] = content
        self.mtiFeatures = features

    def buildTables(self):
        for tag, features in self.mtiFeatures.items():
            table = mtiLib.build(features.splitlines(), self.ttFont)
            assert table.tableTag == tag
            self.ttFont[tag] = table


def warn_about_miscased_insertion_markers(
    ufo_description: str, feaFile: ast.FeatureFile, patterns: set[str]
) -> None:
    """Warn the user about potentially mistyped feature insertion markers."""

    patterns_compiled = tuple(
        (re.compile(pattern), re.compile(pattern, re.IGNORECASE))
        for pattern in patterns
    )

    # NOTE: Insertion markers can only meaningfully occur in top-level feature
    # blocks.
    for block in ast.iterFeatureBlocks(feaFile):
        for statement in block.statements:
            if not isinstance(statement, ast.Comment):
                continue
            for pattern_case, pattern_ignore_case in patterns_compiled:
                text = str(statement)
                match_case = re.match(pattern_case, text)
                match_ignore_case = re.match(pattern_ignore_case, text)
                if match_ignore_case and not match_case:
                    logger.warning(
                        "%s: The insertion comment '%s' in the feature file is "
                        "miscased (search pattern: %s), ignoring it.",
                        ufo_description,
                        text,
                        pattern_case.pattern,
                    )


class VariableFeatureCompiler(FeatureCompiler):
    """Generate a variable feature file and compile OpenType tables from a
    designspace file.
    """

    def __init__(
        self,
        ufo,
        designspace,
        ttFont=None,
        glyphSet=None,
        featureWriters=None,
        **kwargs,
    ):
        self.designspace = designspace
        super().__init__(ufo, ttFont, glyphSet, featureWriters, **kwargs)

    def setupFeatures(self):
        if self.featureWriters:
            featureFile = parseLayoutFeatures(self.ufo, self.feaIncludeDir)

            for writer in self.featureWriters:
                writer.write(self.designspace, featureFile, compiler=self)

            # stringify AST to get correct line numbers in error messages
            self.features = featureFile.asFea()
        else:
            # no featureWriters, simply read existing features' text
            self.features = self.ufo.features.text or ""


def _featuresCompatible(designSpaceDoc: DesignSpaceDocument) -> bool:
    """Returns True when the features of the individual source UFOs are the same,
    or when only the default source has features.

    NOTE: Only compares the feature file text inside the source UFO and does not
    follow imports. This will suffice as long as no external feature file is
    using variable syntax and all sources are stored n the same parent folder
    (so the same includes point to the same files).
    """

    assert all(hasattr(source.font, "features") for source in designSpaceDoc.sources)

    def transform(f: SourceDescriptor) -> str:
        # Strip comments
        text = re.sub("(?m)#.*$", "", f.font.features.text or "")
        # Strip extraneous whitespace
        text = re.sub(r"\s+", " ", text)
        return text

    sources = sorted(
        designSpaceDoc.sources, key=lambda source: source != designSpaceDoc.default
    )
    assert sources[0] == designSpaceDoc.default

    transformed = [transform(s) for s in sources]
    first = transformed[0]
    return all(s == first for s in transformed[1:]) or all(
        not s for s in transformed[1:]
    )



import logging
from textwrap import dedent

import pytest

from ufo2ft.featureCompiler import parseLayoutFeatures
from ufo2ft.featureWriters import GdefFeatureWriter

from . import FeatureWriterTest


@pytest.fixture
def testufo(FontClass):
    ufo = FontClass()
    ufo.newGlyph("a")
    ufo.newGlyph("f")
    ufo.newGlyph("f.component")
    ufo.newGlyph("i")
    liga = ufo.newGlyph("f_f_i")
    liga.appendAnchor({"name": "caret_2", "x": 400, "y": 0})
    liga.appendAnchor({"name": "caret_1", "x": 200, "y": 0})
    liga = ufo.newGlyph("f_i")
    liga.appendAnchor({"name": "caret_", "x": 200, "y": 0})
    ufo.newGlyph("acutecomb")
    ufo.newGlyph("tildecomb")

    ufo.lib["public.glyphOrder"] = [
        "a",
        "f",
        "f.component",
        "i",
        "f_f_i",
        "f_i",
        "acutecomb",
        "tildecomb",
    ]
    return ufo


class GdefFeatureWriterTest(FeatureWriterTest):
    FeatureWriter = GdefFeatureWriter

    @classmethod
    def writeGDEF(cls, ufo, **kwargs):
        writer = cls.FeatureWriter(**kwargs)
        feaFile = parseLayoutFeatures(ufo)
        if writer.write(ufo, feaFile):
            return feaFile

    def test_no_GDEF_no_openTypeCategories_in_font(self, testufo):
        newFea = self.writeGDEF(testufo)
        assert str(newFea) == dedent(
            """\
            table GDEF {
                LigatureCaretByPos f_f_i 200 400;
                LigatureCaretByPos f_i 200;
            } GDEF;
            """
        )

    def test_GDEF_in_font(self, testufo):
        testufo.features.text = dedent(
            """\
            table GDEF {
                GlyphClassDef [a], [], [acutecomb], [];
                LigatureCaretByPos f_i 300;
            } GDEF;
            """
        )
        assert self.writeGDEF(testufo) is None

    def test_openTypeCategories_in_font(self, testufo):
        testufo.lib["public.openTypeCategories"] = {
            "a": "base",
            "f.component": "component",
            "f_i": "ligature",
            "acutecomb": "mark",
        }
        newFea = self.writeGDEF(testufo)
        assert str(newFea) == dedent(
            """\
            table GDEF {
                GlyphClassDef [a], [f_i], [acutecomb], [f.component];
                LigatureCaretByPos f_f_i 200 400;
                LigatureCaretByPos f_i 200;
            } GDEF;
            """
        )

    def test_GDEF_and_openTypeCategories_in_font(self, testufo):
        testufo.lib["public.openTypeCategories"] = {
            "a": "base",
            "f.component": "component",
            "f_i": "ligature",
            "acutecomb": "mark",
        }
        testufo.features.text = dedent(
            """\
            table GDEF {
                 GlyphClassDef [i], [], [tildecomb], [];
                 LigatureCaretByPos f_i 100;
             } GDEF;
             """
        )
        assert self.writeGDEF(testufo) is None

    def test_GDEF_LigatureCarets_and_openTypeCategories_in_font(self, testufo):
        testufo.lib["public.openTypeCategories"] = {
            "a": "base",
            "f.component": "component",
            "f_i": "ligature",
            "acutecomb": "mark",
        }
        testufo.features.text = dedent(
            """\
            table GDEF {
                LigatureCaretByPos f_i 100;
            } GDEF;
             """
        )
        newFea = self.writeGDEF(testufo)
        assert str(newFea) == dedent(
            """\
            table GDEF {
                LigatureCaretByPos f_i 100;
                GlyphClassDef [a], [f_i], [acutecomb], [f.component];
            } GDEF;
            """
        )

    def test_GDEF_GlyphClassDef_and_carets_in_font(self, testufo):
        testufo.lib["public.openTypeCategories"] = {
            "a": "base",
            "f.component": "component",
            "f_i": "ligature",
            "acutecomb": "mark",
        }
        testufo.features.text = dedent(
            """\
            table GDEF {
                GlyphClassDef [], [], [acutecomb tildecomb], [];
            } GDEF;
             """
        )
        newFea = self.writeGDEF(testufo)
        assert str(newFea) == dedent(
            """\
            table GDEF {
                GlyphClassDef [], [], [acutecomb tildecomb], [];
                LigatureCaretByPos f_f_i 200 400;
                LigatureCaretByPos f_i 200;
            } GDEF;
            """
        )

    def test_mark_and_openTypeCategories_in_font(self, testufo):
        testufo.lib["public.openTypeCategories"] = {
            "a": "base",
            "f.component": "component",
            "f_f_i": "base",
            "f_i": "ligature",
            "acutecomb": "mark",
            "tildecomb": "component",
        }
        testufo.features.text = old = dedent(
            """\
            feature mark {
                markClass tildecomb <anchor 0 500> @TOP_MARKS;
                pos base a
                    <anchor 250 500> mark @TOP_MARKS;
                pos base f
                    <anchor 250 500> mark @TOP_MARKS;
                pos ligature f_f_i
                        <anchor 150 700> mark @TOP_MARKS
                    ligComponent
                        <anchor 450 700> mark @TOP_MARKS
                    ligComponent
                        <anchor 600 700> mark @TOP_MARKS;
            } mark;
            """
        )
        newFea = self.writeGDEF(testufo)
        assert str(newFea) == old + "\n" + dedent(
            """\
            table GDEF {
                GlyphClassDef [a f_f_i], [f_i], [acutecomb], [f.component tildecomb];
                LigatureCaretByPos f_f_i 200 400;
                LigatureCaretByPos f_i 200;
            } GDEF;
            """
        )

    def test_vertical_carets(self, testufo):
        vliga = testufo.newGlyph("vi_li_ga")
        vliga.appendAnchor({"name": "vcaret_1", "x": 0, "y": 100})
        vliga.appendAnchor({"name": "vcaret_2", "x": 0, "y": 200})
        vliga = testufo.newGlyph("vli_ga")
        vliga.appendAnchor({"name": "vcaret_", "x": 0, "y": 100})

        newFea = self.writeGDEF(testufo)
        assert str(newFea) == dedent(
            """\
            table GDEF {
                LigatureCaretByPos f_f_i 200 400;
                LigatureCaretByPos f_i 200;
                LigatureCaretByPos vi_li_ga 100 200;
                LigatureCaretByPos vli_ga 100;
            } GDEF;
            """
        )

    def test_floaty_carets(self, testufo):
        # Some Glyphs sources happen to contain fractional caret positions.
        # In the Adobe feature file syntax (and binary OpenType GDEF tables),
        # caret positions must be integers.
        liga = testufo.newGlyph("li_ga")
        liga.appendAnchor({"name": "vcaret_1", "x": 0, "y": 200.1111})
        liga.appendAnchor({"name": "caret_1", "x": 499.9876, "y": 0})

        newFea = self.writeGDEF(testufo)
        assert str(newFea) == dedent(
            """\
            table GDEF {
                LigatureCaretByPos f_f_i 200 400;
                LigatureCaretByPos f_i 200;
                LigatureCaretByPos li_ga 200 500;
            } GDEF;
            """
        )

    def test_getOpenTypeCategories_invalid(self, testufo, caplog):
        caplog.set_level(logging.WARNING)
        testufo.lib["public.openTypeCategories"] = {
            "a": "base",
            "f.component": "component",
            "f_f_i": "base",
            "f_i": "ligature",
            "acutecomb": "mark",
            "tildecomb": "components",
        }
        logger = "ufo2ft.featureWriters.gdefFeatureWriter.GdefFeatureWriter"
        with caplog.at_level(logging.WARNING, logger=logger):
            self.writeGDEF(testufo)

        assert len(caplog.records) == 1
        assert "The 'public.openTypeCategories' value of tildecomb in" in caplog.text
        assert "is 'components' when it should be" in caplog.text



import logging

import fontTools.feaLib.ast as fea_ast
import pytest
from fontTools import unicodedata
from syrupy.extensions.amber import AmberSnapshotExtension
from syrupy.location import PyTestLocation
from syrupy.types import SnapshotIndex

from ufo2ft.constants import UNICODE_SCRIPT_ALIASES
from ufo2ft.featureCompiler import FeatureCompiler, parseLayoutFeatures
from ufo2ft.featureWriters.kernFeatureWriter2 import KernFeatureWriter
from ufo2ft.util import DFLT_SCRIPTS, unicodeScriptExtensions

from . import FeatureWriterTest


class KernFeatureWriterTest(FeatureWriterTest):
    FeatureWriter = KernFeatureWriter


class SameUfoLibResultsExtension(AmberSnapshotExtension):
    """Make tests use the same snapshots when parameterized.

    Instead of having the snapshots of "test_something[defcon]" and
    "test_something[ufoLib2]" be duplicates, use the same snapshots for both,
    because the UFO library shouldn't make a difference.
    """

    @classmethod
    def get_snapshot_name(
        cls, *, test_location: "PyTestLocation", index: "SnapshotIndex"
    ) -> str:
        index_suffix = ""
        if isinstance(index, (str,)):
            index_suffix = f"[{index}]"
        elif index:
            index_suffix = f".{index}"
        return f"{test_location.methodname}{index_suffix}"


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(SameUfoLibResultsExtension)


def makeUFO(cls, glyphMap, groups=None, kerning=None, features=None):
    ufo = cls()
    for name, uni in glyphMap.items():
        glyph = ufo.newGlyph(name)
        if isinstance(uni, (list, tuple)):
            glyph.unicodes = uni
        elif uni is not None:
            glyph.unicode = uni
    if groups is not None:
        ufo.groups.update(groups)
    if kerning is not None:
        ufo.kerning.update(kerning)
    if features is not None:
        ufo.features.text = features
    return ufo


def getClassDefs(feaFile):
    return [
        s for s in feaFile.statements if isinstance(s, fea_ast.GlyphClassDefinition)
    ]


def getGlyphs(classDef):
    return [str(g) for g in classDef.glyphs.glyphSet()]


def getLookups(feaFile):
    return [s for s in feaFile.statements if isinstance(s, fea_ast.LookupBlock)]


def getPairPosRules(lookup):
    return [s for s in lookup.statements if isinstance(s, fea_ast.PairPosStatement)]


def test_cleanup_missing_glyphs(FontClass):
    groups = {
        "public.kern1.A": ["A", "Aacute", "Abreve", "Acircumflex"],
        "public.kern2.B": ["B", "D", "E", "F"],
        "public.kern1.C": ["foobar"],
    }
    kerning = {
        ("public.kern1.A", "public.kern2.B"): 10,
        ("public.kern1.A", "baz"): -25,
        ("baz", "public.kern2.B"): -20,
        ("public.kern1.C", "public.kern2.B"): 20,
    }
    ufo = FontClass()
    exclude = {"Abreve", "D", "foobar"}
    for glyphs in groups.values():
        for glyph in glyphs:
            if glyph in exclude:
                continue
            ufo.newGlyph(glyph)
    ufo.groups.update(groups)
    ufo.kerning.update(kerning)

    writer = KernFeatureWriter()
    feaFile = parseLayoutFeatures(ufo)
    writer.write(ufo, feaFile)

    classDefs = getClassDefs(feaFile)
    assert len(classDefs) == 2
    assert classDefs[0].name == "kern1.dflt.A"
    assert classDefs[1].name == "kern2.dflt.B"
    assert getGlyphs(classDefs[0]) == ["A", "Aacute", "Acircumflex"]
    assert getGlyphs(classDefs[1]) == ["B", "E", "F"]

    lookups = getLookups(feaFile)
    assert len(lookups) == 1
    kern_lookup = lookups[0]
    # We have no codepoints defined for these, so they're considered common
    assert kern_lookup.name == "kern_dflt"
    rules = getPairPosRules(kern_lookup)
    assert len(rules) == 1
    assert str(rules[0]) == "pos @kern1.dflt.A @kern2.dflt.B 10;"


def test_ignoreMarks(snapshot, FontClass):
    font = FontClass()
    for name in ("one", "four", "six"):
        font.newGlyph(name)
    font.kerning.update({("four", "six"): -55.0, ("one", "six"): -30.0})
    # default is ignoreMarks=True
    writer = KernFeatureWriter()
    feaFile = fea_ast.FeatureFile()
    assert writer.write(font, feaFile)

    assert feaFile.asFea() == snapshot

    writer = KernFeatureWriter(ignoreMarks=False)
    feaFile = fea_ast.FeatureFile()
    assert writer.write(font, feaFile)

    assert feaFile.asFea() == snapshot


def test_mark_to_base_kern(snapshot, FontClass):
    font = FontClass()
    for name in ("A", "B", "C"):
        font.newGlyph(name).unicode = ord(name)
    font.newGlyph("acutecomb").unicode = 0x0301
    font.kerning.update({("A", "acutecomb"): -55.0, ("B", "C"): -30.0})

    font.features.text = """\
        @Bases = [A B C];
        @Marks = [acutecomb];
        table GDEF {
            GlyphClassDef @Bases, [], @Marks, ;
        } GDEF;
        """

    # default is ignoreMarks=True
    feaFile = KernFeatureWriterTest.writeFeatures(font)
    assert feaFile.asFea() == snapshot

    feaFile = KernFeatureWriterTest.writeFeatures(font, ignoreMarks=False)
    assert feaFile.asFea() == snapshot


def test_mark_to_base_only(snapshot, FontClass):
    font = FontClass()
    for name in ("A", "B", "C"):
        font.newGlyph(name)
    font.newGlyph("acutecomb").unicode = 0x0301
    font.kerning.update({("A", "acutecomb"): -55.0})

    font.features.text = """\
        @Bases = [A B C];
        @Marks = [acutecomb];
        table GDEF {
            GlyphClassDef @Bases, [], @Marks, ;
        } GDEF;
        """

    # default is ignoreMarks=True
    feaFile = KernFeatureWriterTest.writeFeatures(font)
    assert feaFile.asFea() == snapshot


def test_mode(snapshot, FontClass):
    ufo = FontClass()
    for name in ("one", "four", "six", "seven"):
        ufo.newGlyph(name)
    existing = """\
        feature kern {
            pos one four' -50 six;
        } kern;
        """
    ufo.features.text = existing
    ufo.kerning.update({("seven", "six"): 25.0})

    writer = KernFeatureWriter()  # default mode="skip"
    feaFile = parseLayoutFeatures(ufo)
    assert not writer.write(ufo, feaFile)

    assert str(feaFile) == snapshot(name="existing")

    # pass optional "append" mode
    writer = KernFeatureWriter(mode="append")
    feaFile = parseLayoutFeatures(ufo)
    assert writer.write(ufo, feaFile)

    assert feaFile.asFea() == snapshot

    # pass "skip" mode explicitly
    writer = KernFeatureWriter(mode="skip")
    feaFile = parseLayoutFeatures(ufo)
    assert not writer.write(ufo, feaFile)

    assert feaFile.asFea() == snapshot(name="existing")


def test_insert_comment_before(snapshot, FontClass):
    ufo = FontClass()
    for name in ("one", "four", "six", "seven"):
        ufo.newGlyph(name)
    existing = """\
        feature kern {
            #
            # Automatic Code
            #
            pos one four' -50 six;
        } kern;
        """
    ufo.features.text = existing
    ufo.kerning.update({("seven", "six"): 25.0})

    writer = KernFeatureWriter()
    feaFile = parseLayoutFeatures(ufo)
    assert writer.write(ufo, feaFile)

    assert feaFile.asFea() == snapshot

    # test append mode ignores insert marker
    generated = KernFeatureWriterTest.writeFeatures(ufo, mode="append")
    assert generated.asFea() == snapshot


def test_comment_wrong_case_or_missing(snapshot, FontClass, caplog):
    ufo = FontClass()
    for name in ("a", "b"):
        ufo.newGlyph(name)
    ufo.kerning.update({("a", "b"): 25.0})
    ufo.features.text = (
        """
        feature kern {
            # Automatic code
        } kern;
        """
    ).strip()

    with caplog.at_level(logging.WARNING):
        compiler = FeatureCompiler(ufo, featureWriters=[KernFeatureWriter])
        font = compiler.compile()

    # We mis-cased the insertion marker above, so it's ignored and we end up
    # with an empty `kern` block overriding the other kerning in the font
    # source and therefore no `GPOS` table.
    assert "miscased" in caplog.text
    assert "Dropping the former" in caplog.text
    assert "GPOS" not in font

    # Append mode ignores insertion markers and so should not log warnings
    # and have kerning in the final font.
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        compiler = FeatureCompiler(
            ufo, featureWriters=[KernFeatureWriter(mode="append")]
        )
        font = compiler.compile()

    assert not caplog.text
    assert "GPOS" in font


def test_insert_comment_before_extended(snapshot, FontClass):
    ufo = FontClass()
    for name in ("one", "four", "six", "seven"):
        ufo.newGlyph(name)
    existing = """\
        feature kern {
            #
            # Automatic Code End
            #
            pos one four' -50 six;
        } kern;
        """
    ufo.features.text = existing
    ufo.kerning.update({("seven", "six"): 25.0})

    writer = KernFeatureWriter()
    feaFile = parseLayoutFeatures(ufo)
    assert writer.write(ufo, feaFile)

    assert feaFile.asFea() == snapshot


def test_insert_comment_after(snapshot, FontClass):
    ufo = FontClass()
    for name in ("one", "four", "six", "seven"):
        ufo.newGlyph(name)
    existing = """\
        feature kern {
            pos one four' -50 six;
            #
            # Automatic Code
            #
        } kern;
        """
    ufo.features.text = existing
    ufo.kerning.update({("seven", "six"): 25.0})

    writer = KernFeatureWriter()
    feaFile = parseLayoutFeatures(ufo)
    assert writer.write(ufo, feaFile)

    assert feaFile.asFea() == snapshot

    # test append mode ignores insert marker
    generated = KernFeatureWriterTest.writeFeatures(ufo, mode="append")
    assert generated.asFea() == snapshot


def test_insert_comment_middle(snapshot, FontClass):
    ufo = FontClass()
    for name in ("one", "four", "six", "seven"):
        ufo.newGlyph(name)
    existing = """\
        feature kern {
            pos one four' -50 six;
            #
            # Automatic Code
            #
            pos one six' -50 six;
        } kern;
        """
    ufo.features.text = existing
    ufo.kerning.update({("seven", "six"): 25.0})

    writer = KernFeatureWriter()
    feaFile = parseLayoutFeatures(ufo)

    writer.write(ufo, feaFile)
    assert str(feaFile) == snapshot

    # test append mode ignores insert marker
    generated = KernFeatureWriterTest.writeFeatures(ufo, mode="append")
    assert generated.asFea() == snapshot


def test_arabic_numerals(snapshot, FontClass):
    """Test that arabic numerals (with bidi type AN) are kerned LTR.

    See:

    * https://github.com/googlei18n/ufo2ft/issues/198
    * https://github.com/googlei18n/ufo2ft/pull/200

    Additionally, some Arabic numerals are used in more than one script. One
    approach is to look at other glyphs with distinct script associations
    and consider the font to be supporting those.
    """
    ufo = FontClass()
    for name, code in [("four-ar", 0x664), ("seven-ar", 0x667)]:
        glyph = ufo.newGlyph(name)
        glyph.unicode = code
    ufo.kerning.update({("four-ar", "seven-ar"): -30})

    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot

    ufo.newGlyph("alef-ar").unicode = 0x627
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot

    ufo.features.text = """
        languagesystem DFLT dflt;
        languagesystem Thaa dflt;
    """
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot

    del ufo["alef-ar"]
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot


def test_skip_zero_class_kerns(snapshot, FontClass):
    glyphs = {
        "A": ord("A"),
        "B": ord("B"),
        "C": ord("C"),
        "D": ord("D"),
        "E": ord("E"),
        "F": ord("F"),
        "G": ord("G"),
        "H": ord("H"),
    }
    groups = {
        "public.kern1.foo": ["A", "B"],
        "public.kern2.bar": ["C", "D"],
        "public.kern1.baz": ["E", "F"],
        "public.kern2.nul": ["G", "H"],
    }
    kerning = {
        ("public.kern1.foo", "public.kern2.bar"): 10,
        ("public.kern1.baz", "public.kern2.bar"): -10,
        ("public.kern1.foo", "D"): 15,
        ("A", "public.kern2.bar"): 5,
        ("G", "H"): -5,
        # class-class zero-value pairs are skipped
        ("public.kern1.foo", "public.kern2.nul"): 0,
    }

    ufo = makeUFO(FontClass, glyphs, groups, kerning)
    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


def test_kern_uniqueness(snapshot, FontClass):
    glyphs = {
        ".notdef": None,
        "questiondown": 0xBF,
        "y": 0x79,
    }
    groups = {
        "public.kern1.questiondown": ["questiondown"],
        "public.kern2.y": ["y"],
    }
    kerning = {
        ("public.kern1.questiondown", "public.kern2.y"): 15,
        ("public.kern1.questiondown", "y"): 35,
        ("questiondown", "public.kern2.y"): -35,
        ("questiondown", "y"): 35,
    }
    ufo = makeUFO(FontClass, glyphs, groups, kerning)

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    # The final kerning value for questiondown, y is 35 and all variants
    # must be present. Ensures the uniqueness filter doesn't filter things
    # out.
    assert newFeatures.asFea() == snapshot


def test_kern_LTR_and_RTL(snapshot, FontClass):
    glyphs = {
        ".notdef": None,
        "four": 0x34,
        "seven": 0x37,
        "A": 0x41,
        "V": 0x56,
        "Aacute": 0xC1,
        "alef-ar": 0x627,
        "reh-ar": 0x631,
        "zain-ar": 0x632,
        "lam-ar": 0x644,
        "four-ar": 0x664,
        "seven-ar": 0x667,
        # # we also add glyphs without unicode codepoint, but linked to
        # # an encoded 'character' glyph by some GSUB rule
        "alef-ar.isol": None,
        "lam-ar.init": None,
        "reh-ar.fina": None,
    }
    groups = {
        "public.kern1.A": ["A", "Aacute"],
        "public.kern1.reh": ["reh-ar", "zain-ar", "reh-ar.fina"],
        "public.kern2.alef": ["alef-ar", "alef-ar.isol"],
    }
    kerning = {
        ("public.kern1.A", "V"): -40,
        ("seven", "four"): -25,
        ("reh-ar.fina", "lam-ar.init"): -80,
        ("public.kern1.reh", "public.kern2.alef"): -100,
        ("four-ar", "seven-ar"): -30,
    }
    features = """\
        languagesystem DFLT dflt;
        languagesystem latn dflt;
        languagesystem latn TRK;
        languagesystem arab dflt;
        languagesystem arab URD;

        feature init {
            script arab;
            sub lam-ar by lam-ar.init;
            language URD;
        } init;

        feature fina {
            script arab;
            sub reh-ar by reh-ar.fina;
            language URD;
        } fina;

        feature isol {
            script arab;
            sub alef-ar by alef-ar.isol;
        } isol;
        """

    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo, ignoreMarks=False)

    assert newFeatures.asFea() == snapshot


def test_kern_LTR_and_RTL_with_marks(snapshot, FontClass):
    glyphs = {
        ".notdef": None,
        "four": 0x34,
        "seven": 0x37,
        "A": 0x41,
        "V": 0x56,
        "Aacute": 0xC1,
        "acutecomb": 0x301,
        "alef-ar": 0x627,
        "reh-ar": 0x631,
        "zain-ar": 0x632,
        "lam-ar": 0x644,
        "four-ar": 0x664,
        "seven-ar": 0x667,
        "fatha-ar": 0x64E,
        # we also add glyphs without unicode codepoint, but linked to
        # an encoded 'character' glyph by some GSUB rule
        "alef-ar.isol": None,
        "lam-ar.init": None,
        "reh-ar.fina": None,
    }
    groups = {
        "public.kern1.A": ["A", "Aacute"],
        "public.kern1.reh": ["reh-ar", "zain-ar", "reh-ar.fina"],
        "public.kern2.alef": ["alef-ar", "alef-ar.isol"],
    }
    kerning = {
        ("public.kern1.A", "V"): -40,
        ("seven", "four"): -25,
        ("reh-ar.fina", "lam-ar.init"): -80,
        ("public.kern1.reh", "public.kern2.alef"): -100,
        ("four-ar", "seven-ar"): -30,
        ("V", "acutecomb"): 70,
        ("reh-ar", "fatha-ar"): 80,
    }
    features = """\
        languagesystem DFLT dflt;
        languagesystem latn dflt;
        languagesystem latn TRK;
        languagesystem arab dflt;
        languagesystem arab URD;

        feature init {
            script arab;
            sub lam-ar by lam-ar.init;
            language URD;
        } init;

        feature fina {
            script arab;
            sub reh-ar by reh-ar.fina;
            language URD;
        } fina;

        feature isol {
            script arab;
            sub alef-ar by alef-ar.isol;
        } isol;

        @Bases = [A V Aacute alef-ar reh-ar zain-ar lam-ar
                    alef-ar.isol lam-ar.init reh-ar.fina];
        @Marks = [acutecomb fatha-ar];
        table GDEF {
            GlyphClassDef @Bases, [], @Marks, ;
        } GDEF;
        """

    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


def test_kern_RTL_with_marks(snapshot, FontClass):
    glyphs = {
        ".notdef": None,
        "alef-ar": 0x627,
        "reh-ar": 0x631,
        "zain-ar": 0x632,
        "lam-ar": 0x644,
        "four-ar": 0x664,
        "seven-ar": 0x667,
        "fatha-ar": 0x64E,
        # we also add glyphs without unicode codepoint, but linked to
        # an encoded 'character' glyph by some GSUB rule
        "alef-ar.isol": None,
        "lam-ar.init": None,
        "reh-ar.fina": None,
    }
    groups = {
        "public.kern1.reh": ["reh-ar", "zain-ar", "reh-ar.fina"],
        "public.kern2.alef": ["alef-ar", "alef-ar.isol"],
    }
    kerning = {
        ("reh-ar.fina", "lam-ar.init"): -80,
        ("public.kern1.reh", "public.kern2.alef"): -100,
        ("reh-ar", "fatha-ar"): 80,
    }
    features = """\
        languagesystem arab dflt;
        languagesystem arab ARA;

        feature init {
            script arab;
            sub lam-ar by lam-ar.init;
        } init;

        feature fina {
            script arab;
            sub reh-ar by reh-ar.fina;
        } fina;

        feature isol {
            script arab;
            sub alef-ar by alef-ar.isol;
        } isol;

        @Bases = [alef-ar reh-ar zain-ar lam-ar alef-ar.isol lam-ar.init reh-ar.fina];
        @Marks = [fatha-ar];
        table GDEF {
            GlyphClassDef @Bases, [], @Marks, ;
        } GDEF;
        """

    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)
    assert newFeatures.asFea() == snapshot


def test_kern_independent_of_languagesystem(snapshot, FontClass):
    glyphs = {"A": 0x41, "V": 0x56, "reh-ar": 0x631, "alef-ar": 0x627}
    kerning = {("A", "V"): -40, ("reh-ar", "alef-ar"): -100}
    # No languagesystems declared.
    ufo = makeUFO(FontClass, glyphs, kerning=kerning)
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot(name="same")

    features = "languagesystem arab dflt;"
    ufo = makeUFO(FontClass, glyphs, kerning=kerning, features=features)
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot(name="same")


def test_dist_LTR(snapshot, FontClass):
    glyphs = {"aaMatra_kannada": 0x0CBE, "ailength_kannada": 0xCD6}
    groups = {
        "public.kern1.KND_aaMatra_R": ["aaMatra_kannada"],
        "public.kern2.KND_ailength_L": ["aaMatra_kannada"],
    }
    kerning = {("public.kern1.KND_aaMatra_R", "public.kern2.KND_ailength_L"): 34}
    features = """\
        languagesystem DFLT dflt;
        languagesystem latn dflt;
        languagesystem knda dflt;
        languagesystem knd2 dflt;
        """

    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot


def test_dist_RTL(snapshot, FontClass):
    glyphs = {"u10A06": 0x10A06, "u10A1E": 0x10A1E}
    kerning = {("u10A1E", "u10A06"): 117}
    features = """\
        languagesystem DFLT dflt;
        languagesystem arab dflt;
        languagesystem khar dflt;
        """

    ufo = makeUFO(FontClass, glyphs, kerning=kerning, features=features)
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot


def test_dist_LTR_and_RTL(snapshot, FontClass):
    glyphs = {
        "aaMatra_kannada": 0x0CBE,
        "ailength_kannada": 0xCD6,
        "u10A06": 0x10A06,
        "u10A1E": 0x10A1E,
    }
    groups = {
        "public.kern1.KND_aaMatra_R": ["aaMatra_kannada"],
        "public.kern2.KND_ailength_L": ["aaMatra_kannada"],
    }
    kerning = {
        ("public.kern1.KND_aaMatra_R", "public.kern2.KND_ailength_L"): 34,
        ("u10A1E", "u10A06"): 117,
    }
    features = """\
            languagesystem DFLT dflt;
            languagesystem knda dflt;
            languagesystem knd2 dflt;
            languagesystem khar dflt;
            """

    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot


def test_ambiguous_direction_pair(snapshot, FontClass, caplog):
    """Test that glyphs with ambiguous BiDi classes get split correctly."""

    glyphs = {
        "A": 0x0041,
        "one": 0x0031,
        "yod-hb": 0x05D9,
        "reh-ar": 0x0631,
        "one-ar": 0x0661,
        "bar": [0x0073, 0x0627],
    }
    kerning = {
        ("bar", "bar"): 1,
        ("bar", "A"): 2,
        ("reh-ar", "A"): 3,
        ("reh-ar", "one-ar"): 4,
        ("yod-hb", "one"): 5,
    }
    features = """\
        languagesystem DFLT dflt;
        languagesystem latn dflt;
        languagesystem arab dflt;
        """
    ufo = makeUFO(FontClass, glyphs, kerning=kerning, features=features)

    with caplog.at_level(logging.INFO):
        generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot
    assert caplog.messages == [
        "Skipping part of a kerning pair <bar bar 1> with mixed direction (LeftToRight, RightToLeft)",  # noqa: B950
        "Skipping part of a kerning pair <bar bar 1> with mixed direction (RightToLeft, LeftToRight)",  # noqa: B950
        "Skipping part of a kerning pair <bar A 2> with conflicting BiDi classes",  # noqa: B950
        "Skipping part of a kerning pair <bar A 2> with mixed direction (RightToLeft, LeftToRight)",  # noqa: B950
        "Skipping part of a kerning pair <reh-ar A 3> with mixed direction (RightToLeft, LeftToRight)",  # noqa: B950
        "Skipping part of a kerning pair <reh-ar one-ar 4> with conflicting BiDi classes",  # noqa: B950
        "Skipping part of a kerning pair <yod-hb one 5> with conflicting BiDi classes",  # noqa: B950
    ]


def test_kern_RTL_and_DFLT_numbers(snapshot, FontClass):
    glyphs = {"four": 0x34, "seven": 0x37, "bet-hb": 0x5D1, "yod-hb": 0x5D9}
    kerning = {("seven", "four"): -25, ("yod-hb", "bet-hb"): -100}
    features = """\
        languagesystem DFLT dflt;
        languagesystem hebr dflt;
        """

    ufo = makeUFO(FontClass, glyphs, kerning=kerning, features=features)
    generated = KernFeatureWriterTest.writeFeatures(ufo)

    assert generated.asFea() == snapshot


def test_quantize(snapshot, FontClass):
    font = FontClass()
    for name in ("one", "four", "six"):
        font.newGlyph(name)
    font.kerning.update({("four", "six"): -57.0, ("one", "six"): -24.0})
    writer = KernFeatureWriter(quantization=5)
    feaFile = fea_ast.FeatureFile()
    writer.write(font, feaFile)

    assert feaFile.asFea() == snapshot


def test_skip_spacing_marks(snapshot, data_dir, FontClass):
    fontPath = data_dir / "SpacingCombiningTest-Regular.ufo"
    testufo = FontClass(fontPath)
    generated = KernFeatureWriterTest.writeFeatures(testufo)

    assert generated.asFea() == snapshot


def test_kern_split_multi_glyph_class(snapshot, FontClass):
    """Test that kern pair types are correctly split across directions."""

    glyphs = {
        "a": ord("a"),
        "b": ord("b"),
        "period": ord("."),
    }
    groups = {
        "public.kern1.foo": ["a", "period"],
        "public.kern2.foo": ["b", "period"],
    }
    kerning = {
        # Glyph-to-glyph
        ("a", "a"): 1,
        ("a", "b"): 2,
        ("a", "period"): 3,
        ("b", "a"): 4,
        ("b", "b"): 5,
        ("b", "period"): 6,
        ("period", "a"): 7,
        ("period", "b"): 8,
        ("period", "period"): 9,
        # Class-to-glyph
        ("public.kern1.foo", "b"): 10,
        ("public.kern1.foo", "period"): 11,
        # Glyph-to-class
        ("a", "public.kern2.foo"): 12,
        ("period", "public.kern2.foo"): 13,
        # Class-to-class
        ("public.kern1.foo", "public.kern2.foo"): 14,
    }

    ufo = makeUFO(FontClass, glyphs, groups, kerning)
    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot(name="same")

    # Making a common glyph implicitly have an explicit script assigned (GSUB
    # closure) will still keep it in the common section.
    features = """
        feature ss01 {
            sub a by period; # Make period be both Latn and Zyyy.
        } ss01;
        """

    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)
    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot(name="same")


def test_kern_split_and_drop(snapshot, FontClass, caplog):
    """Test that mixed directions pairs are pruned and only the compatible parts
    are kept."""

    glyphs = {
        "a": ord("a"),
        "alpha": ord("α"),
        "a-orya": 0x0B05,
        "a-cy": 0x0430,
        "alef-ar": 0x627,
        "period": ord("."),
    }
    groups = {
        "public.kern1.foo": ["a", "alpha", "a-orya"],
        "public.kern2.foo": ["a", "alpha", "a-orya"],
        "public.kern1.bar": ["a-cy", "alef-ar", "period"],
        "public.kern2.bar": ["a-cy", "alef-ar", "period"],
    }
    kerning = {
        ("public.kern1.foo", "public.kern2.bar"): 20,
        ("public.kern1.bar", "public.kern2.foo"): 20,
    }

    ufo = makeUFO(FontClass, glyphs, groups, kerning)
    with caplog.at_level(logging.INFO):
        newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot
    assert caplog.messages == [
        "Skipping part of a kerning pair <('a', 'a-orya', 'alpha') ('alef-ar',) 20> with mixed direction (LeftToRight, RightToLeft)",  # noqa: B950
        "Skipping part of a kerning pair <('alef-ar',) ('a', 'a-orya', 'alpha') 20> with mixed direction (RightToLeft, LeftToRight)",  # noqa: B950
    ]


def test_kern_split_and_drop_mixed(snapshot, caplog, FontClass):
    """Test that mixed directions pairs are dropped.

    And that scripts with no remaining lookups don't crash.
    """

    glyphs = {"V": ord("V"), "W": ord("W"), "gba-nko": 0x07DC}
    groups = {"public.kern1.foo": ["V", "W"], "public.kern2.foo": ["gba-nko", "W"]}
    kerning = {("public.kern1.foo", "public.kern2.foo"): -20}
    ufo = makeUFO(FontClass, glyphs, groups, kerning)
    with caplog.at_level(logging.INFO):
        newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot
    assert (
        "<('V', 'W') ('gba-nko',) -20> with mixed direction (LeftToRight, RightToLeft)"
        in caplog.text
    )


def test_kern_mixed_bidis(snapshot, caplog, FontClass):
    """Test that BiDi types for pairs are respected."""

    # TODO: Add Adlam numbers (rtl)
    glyphs = {
        "a": ord("a"),
        "comma": ord(","),
        "alef-ar": 0x0627,
        "comma-ar": 0x060C,
        "one-ar": 0x0661,
        "one-adlam": 0x1E951,
    }
    kerning = {
        # Undetermined: LTR
        ("comma", "comma"): -1,
        # LTR
        ("a", "a"): 1,
        ("a", "comma"): 2,
        ("comma", "a"): 3,
        # RTL
        ("alef-ar", "alef-ar"): 4,
        ("alef-ar", "comma-ar"): 5,
        ("comma-ar", "alef-ar"): 6,
        ("one-adlam", "one-adlam"): 10,
        ("one-adlam", "comma-ar"): 11,
        ("comma-ar", "one-adlam"): 12,
        # Mixed: should be dropped
        ("alef-ar", "one-ar"): 7,
        ("one-ar", "alef-ar"): 8,
        ("one-ar", "one-adlam"): 13,
        # LTR despite being an RTL script
        ("one-ar", "one-ar"): 9,
    }
    ufo = makeUFO(FontClass, glyphs, None, kerning)
    with caplog.at_level(logging.INFO):
        newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot
    assert "<alef-ar one-ar 7> with conflicting BiDi classes" in caplog.text
    assert "<one-ar alef-ar 8> with conflicting BiDi classes" in caplog.text
    assert "<one-ar one-adlam 13> with conflicting BiDi classes" in caplog.text


def unicodeScript(codepoint: int) -> str:
    """Returns the Unicode script for a codepoint, combining some
    scripts into the same bucket.

    This allows lookups to contain more than one script. The most prominent case
    is being able to kern Hiragana and Katakana against each other, Unicode
    defines "Hrkt" as an alias for both scripts.

    Note: Keep in sync with unicodeScriptExtensions!
    """
    script = unicodedata.script(chr(codepoint))
    return UNICODE_SCRIPT_ALIASES.get(script, script)


def test_kern_zyyy_zinh(snapshot, FontClass):
    """Test that a sampling of glyphs with a common or inherited script, but a
    disjoint set of explicit script extensions end up in the correct lookups."""
    glyphs = {}
    for i in range(0, 0x110000, 0x10):
        script = unicodeScript(i)
        script_extension = unicodeScriptExtensions(i)
        if script not in script_extension:
            assert script in DFLT_SCRIPTS
            name = f"uni{i:04X}"
            glyphs[name] = i
    kerning = {(glyph, glyph): i for i, glyph in enumerate(glyphs)}
    ufo = makeUFO(FontClass, glyphs, None, kerning)
    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


def test_kern_hira_kana_hrkt(snapshot, FontClass):
    """Test that Hiragana and Katakana lands in the same lookup and can be
    kerned against each other and common glyphs are kerned just once."""
    glyphs = {"a-hira": 0x3042, "a-kana": 0x30A2, "period": ord(".")}
    kerning = {
        ("a-hira", "a-hira"): 1,
        ("a-hira", "a-kana"): 2,
        ("a-kana", "a-hira"): 3,
        ("a-kana", "a-kana"): 4,
        ("period", "period"): 5,
        ("a-hira", "period"): 6,
        ("period", "a-hira"): 7,
        ("a-kana", "period"): 8,
        ("period", "a-kana"): 9,
    }
    ufo = makeUFO(FontClass, glyphs, None, kerning)
    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


# TODO: Keep? Then modify comments.
def test_defining_classdefs(snapshot, FontClass):
    """Check that we aren't redefining class definitions with different
    content."""

    glyphs = {
        "halant-telugu": 0xC4D,  # Telu
        "ka-telugu.below": None,  # Telu by substitution
        "ka-telugu": 0xC15,  # Telu
        "rVocalicMatra-telugu": 0xC43,  # Telu
        "sha-telugu.below": None,  # Default
        "ss-telugu.alt": None,  # Default
        "ssa-telugu.alt": None,  # Telu by substitution
        "ssa-telugu": 0xC37,  # Telu
    }
    groups = {
        "public.kern1.sha-telugu.below": ["sha-telugu.below"],
        # The following group is a mix of Telu and Default through its gylphs. The
        # kerning for bases below will create a Telu and Default split group.
        # Important for the NOTE below.
        "public.kern1.ssa-telugu.alt": ["ssa-telugu.alt", "ss-telugu.alt"],
        "public.kern2.ka-telugu.below": ["ka-telugu.below"],
        "public.kern2.rVocalicMatra-telugu": ["rVocalicMatra-telugu"],
    }
    kerning = {
        # The follwoing three pairs are base-to-base pairs:
        ("public.kern1.sha-telugu.below", "public.kern2.ka-telugu.below"): 20,
        ("public.kern1.ssa-telugu.alt", "public.kern2.ka-telugu.below"): 60,
        ("public.kern1.ssa-telugu.alt", "sha-telugu.below"): 150,
        # NOTE: This last pair kerns bases against marks, triggering an extra
        # pass to make a mark lookup that will create new classDefs. This extra
        # pass will work on just this one pair, and kern splitting won't split
        # off a Default group from `public.kern1.ssa-telugu.alt`, you get just a
        # Telu pair. Unless the writer keeps track of which classDefs it already
        # generated, this will overwrite the previous `@kern1.Telu.ssatelugu.alt
        # = [ssa-telugu.alt]` with `@kern1.Telu.ssatelugu.alt =
        # [ss-telugu.alt]`, losing kerning.
        ("public.kern1.ssa-telugu.alt", "public.kern2.rVocalicMatra-telugu"): 180,
    }
    features = """
        feature blwf {
            script tel2;
            sub halant-telugu ka-telugu by ka-telugu.below;
        } blwf;

        feature psts {
            script tel2;
            sub ssa-telugu' [rVocalicMatra-telugu sha-telugu.below ka-telugu.below] by ssa-telugu.alt;
        } psts;
    """  # noqa: B950
    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)
    ufo.lib["public.openTypeCategories"] = {
        "halant-telugu": "mark",
        "ka-telugu": "base",
        "rVocalicMatra-telugu": "mark",
        "ss-telugu.alt": "base",
        "ssa-telugu.alt": "base",
        "ssa-telugu": "base",
    }

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


def test_mark_base_kerning(snapshot, FontClass):
    """Check that kerning of bases against marks is correctly split into
    base-only and mixed-mark-and-base lookups, to preserve the semantics of
    kerning exceptions (pairs modifying the effect of other pairs)."""

    glyphs = {
        "aa-tamil": 0x0B86,
        "va-tamil": 0x0BB5,
        "aulengthmark-tamil": 0x0BD7,
    }
    groups = {
        # Each group is a mix of mark and base glyph.
        "public.kern1.e-tamil": ["aulengthmark-tamil", "va-tamil"],
        "public.kern2.e-tamil": ["aulengthmark-tamil", "va-tamil"],
    }
    kerning = {
        ("aa-tamil", "va-tamil"): -20,
        ("aa-tamil", "public.kern2.e-tamil"): -35,
        ("va-tamil", "aa-tamil"): -20,
        ("public.kern1.e-tamil", "aa-tamil"): -35,
        ("aulengthmark-tamil", "aulengthmark-tamil"): -200,
        ("public.kern1.e-tamil", "public.kern2.e-tamil"): -100,
    }
    ufo = makeUFO(FontClass, glyphs, groups, kerning)
    ufo.lib["public.openTypeCategories"] = {
        "aulengthmark-tamil": "mark",
    }

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


def test_hyphenated_duplicates(snapshot, FontClass):
    """Check that kerning group names are kept separate even if their sanitized
    names are the same."""

    glyphs = {"comma": ord(","), "period": ord(".")}
    groups = {
        "public.kern1.hy-phen": ["comma"],
        "public.kern1.hyp-hen": ["period"],
    }
    kerning = {
        ("public.kern1.hy-phen", "comma"): 1,
        ("public.kern1.hyp-hen", "period"): 2,
    }
    ufo = makeUFO(FontClass, glyphs, groups, kerning)

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot


def test_dflt_language(snapshot, FontClass):
    """Check that languages defined for the special DFLT script are registered
    as well."""

    glyphs = {"a": ord("a"), "comma": ord(",")}
    groups = {}
    kerning = {("a", "a"): 1, ("comma", "comma"): 2}
    features = """
        languagesystem DFLT dflt;
        languagesystem DFLT ZND;
        languagesystem latn dflt;
        languagesystem latn ANG;
    """
    ufo = makeUFO(FontClass, glyphs, groups, kerning, features)

    newFeatures = KernFeatureWriterTest.writeFeatures(ufo)

    assert newFeatures.asFea() == snapshot



import logging

import pytest
from fontTools.pens.basePen import MissingComponentError

from ufo2ft.filters.decomposeComponents import (
    DecomposeComponentsFilter,
    DecomposeComponentsIFilter,
)
from ufo2ft.instantiator import Instantiator
from ufo2ft.util import _GlyphSet


def test_missing_component_error(FontClass, caplog):
    ufo = FontClass()
    a = ufo.newGlyph("a")
    a.width = 100
    pen = a.getPen()
    pen.moveTo((0, 0))
    pen.lineTo((300, 0))
    pen.lineTo((300, 300))
    pen.lineTo((0, 300))
    pen.closePath()

    aacute = ufo.newGlyph("aacute")
    aacute.width = 100
    pen = aacute.getPen()
    pen.addComponent("a", (1, 0, 0, 1, 0, 0))
    pen.addComponent("acute", (1, 0, 0, 1, 350, 0))  # missing

    assert len(ufo["aacute"]) == 0
    assert len(ufo["aacute"].components) == 2

    filter_ = DecomposeComponentsFilter()

    with pytest.raises(MissingComponentError, match="'acute'"):
        filter_(ufo)


def test_nested_components(FontClass):
    ufo = FontClass()
    a = ufo.newGlyph("six.lf")
    a.width = 100
    pen = a.getPen()
    pen.moveTo((0, 0))
    pen.lineTo((300, 0))
    pen.lineTo((300, 300))
    pen.lineTo((0, 300))
    pen.closePath()

    b = ufo.newGlyph("nine.lf")
    b.width = 100
    pen = b.getPen()
    pen.addComponent("six.lf", (-1, 0, 0, -1, 0, 0))

    c = ufo.newGlyph("nine")
    c.width = 100
    pen = c.getPen()
    pen.addComponent("nine.lf", (1, 0, 0, 1, 0, 0))

    filter_ = DecomposeComponentsFilter()

    assert filter_(ufo)
    assert len(ufo["six.lf"]) == 1
    assert not ufo["six.lf"].components
    assert len(ufo["nine.lf"]) == 1
    assert not ufo["nine.lf"].components
    assert len(ufo["nine"]) == 1
    assert not ufo["nine"].components


@pytest.fixture
def ufos_and_glyphSets(FontClass):
    """Return two parallel lists of UFOs and glyphSets for testing.

    This fixture creates two UFOs, a Regular and a Bold, each containing 5 glyphs:
    "agrave" composite glyph composed from "a" and "gravecomb" components,
    both in turn simple contour glyphs, and "igrave" composed of "dotlessi" and
    "gravecomb" components.
    The Regular UFO also contains a 'sparse' Medium layer with only two glyphs:
    a different "agrave" composite glyph, but no "a" nor "gravecomb" components;
    and a different shape for the "dotlessi" glyph, but no "igrave" nor "gravecomb".

    The decomposing (interpolatable) filter should interpolate the missing
    components or composites on-the-fly using the instantiator, when available.
    """

    regular_ufo = FontClass()

    a = regular_ufo.newGlyph("a")
    a.width = 500
    a.unicodes = [ord("a")]
    pen = a.getPointPen()
    pen.beginPath()
    pen.addPoint((100, 0), "line")
    pen.addPoint((400, 0), "line")
    pen.addPoint((400, 500), "line")
    pen.addPoint((100, 500), "line")
    pen.endPath()

    i = regular_ufo.newGlyph("dotlessi")
    i.width = 300
    i.unicodes = [0x0131]
    pen = i.getPointPen()
    pen.beginPath()
    pen.addPoint((100, 0), "line")
    pen.addPoint((200, 0), "line")
    pen.addPoint((200, 500), "line")
    pen.addPoint((100, 500), "line")
    pen.endPath()

    gravecomb = regular_ufo.newGlyph("gravecomb")
    gravecomb.unicodes = [0x0300]
    pen = gravecomb.getPointPen()
    pen.beginPath()
    pen.addPoint((30, 550), "line")
    pen.addPoint((0, 750), "line")
    pen.addPoint((-50, 750), "line")
    pen.addPoint((0, 550), "line")
    pen.endPath()

    agrave = regular_ufo.newGlyph("agrave")
    agrave.width = a.width
    agrave.unicodes = [0x00E0]
    pen = agrave.getPointPen()
    pen.addComponent("a", (1, 0, 0, 1, 0, 0))
    pen.addComponent("gravecomb", (1, 0, 0, 1, 250, 0))

    igrave = regular_ufo.newGlyph("igrave")
    igrave.width = i.width
    igrave.unicodes = [0x00EC]
    pen = igrave.getPointPen()
    pen.addComponent("dotlessi", (1, 0, 0, 1, 0, 0))
    pen.addComponent("gravecomb", (1, 0, 0, 1, 150, 0))

    # The Medium layer has "agrave" but does not have "a" and "gravecomb"
    medium_layer = regular_ufo.newLayer("Medium")

    agrave = medium_layer.newGlyph("agrave")
    agrave.width = 550
    pen = agrave.getPointPen()
    pen.addComponent("a", (1, 0, 0, 1, 0, 0))
    pen.addComponent("gravecomb", (1, 0, 0, 1, 275, 0))

    # The Medium layer also has a different "dotlessi" glyph, which the
    # other layers don't have.
    i = medium_layer.newGlyph("dotlessi")
    i.width = 350
    pen = i.getPointPen()
    pen.beginPath()
    pen.addPoint((100, 0), "line")
    pen.addPoint((250, 0), "line")
    pen.addPoint((175, 500), "line")
    pen.addPoint((175, 500), "line")
    pen.endPath()

    bold_ufo = FontClass()

    a = bold_ufo.newGlyph("a")
    a.width = 600
    pen = a.getPointPen()
    pen.beginPath()
    pen.addPoint((150, 0), "line")
    pen.addPoint((450, 0), "line")
    pen.addPoint((450, 500), "line")
    pen.addPoint((150, 500), "line")
    pen.endPath()

    i = bold_ufo.newGlyph("dotlessi")
    i.width = 400
    pen = i.getPointPen()
    pen.beginPath()
    pen.addPoint((100, 0), "line")
    pen.addPoint((300, 0), "line")
    pen.addPoint((300, 500), "line")
    pen.addPoint((100, 500), "line")
    pen.endPath()

    gravecomb = bold_ufo.newGlyph("gravecomb")
    pen = gravecomb.getPointPen()
    pen.beginPath()
    pen.addPoint((40, 550), "line")
    pen.addPoint((0, 750), "line")
    pen.addPoint((-70, 750), "line")
    pen.addPoint((0, 550), "line")
    pen.endPath()

    agrave = bold_ufo.newGlyph("agrave")
    agrave.width = a.width
    pen = agrave.getPointPen()
    pen.addComponent("a", (1, 0, 0, 1, 0, 0))
    pen.addComponent("gravecomb", (1, 0, 0, 1, 300, 0))

    igrave = bold_ufo.newGlyph("igrave")
    igrave.width = i.width
    pen = igrave.getPointPen()
    pen.addComponent("dotlessi", (1, 0, 0, 1, 0, 0))
    pen.addComponent("gravecomb", (1, 0, 0, 1, 200, 0))

    ufos = [regular_ufo, regular_ufo, bold_ufo]
    glyphSets = [
        _GlyphSet.from_layer(regular_ufo),
        _GlyphSet.from_layer(regular_ufo, layerName="Medium"),
        _GlyphSet.from_layer(bold_ufo),
    ]
    return ufos, glyphSets


class DecomposeComponentsIFilterTest:
    def test_composite_with_intermediate_master(self, ufos_and_glyphSets):
        ufos, glyphSets = ufos_and_glyphSets
        regular_glyphs, medium_glyphs, bold_glyphs = glyphSets
        assert "agrave" in medium_glyphs
        assert {"a", "gravecomb"}.isdisjoint(medium_glyphs)

        instantiator = Instantiator(
            {"Weight": (100, 100, 200)},
            [
                ({"Weight": 100}, regular_glyphs),
                ({"Weight": 150}, medium_glyphs),
                ({"Weight": 200}, bold_glyphs),
            ],
        )
        filter_ = DecomposeComponentsIFilter(include={"agrave"})

        modified = filter_(ufos, glyphSets, instantiator=instantiator)

        assert modified == {"agrave"}

        agrave = regular_glyphs["agrave"]
        assert len(agrave.components) == 0
        assert [[(p.x, p.y) for p in c] for c in agrave] == [
            [(100, 0), (400, 0), (400, 500), (100, 500)],
            [(280, 550), (250, 750), (200, 750), (250, 550)],
        ]

        # 'agrave' was fully decomposed also in the medium layer, despite the
        # latter not containing sources for the "a" and "gravecomb" component glyphs.
        # These were interpolated on-the-fly while decomposing the composite glyph.
        agrave = medium_glyphs["agrave"]
        assert len(agrave.components) == 0
        assert [[(p.x, p.y) for p in c] for c in agrave] == [
            [(125, 0), (425, 0), (425, 500), (125, 500)],
            [(310, 550), (275, 750), (215, 750), (275, 550)],
        ]

        agrave = bold_glyphs["agrave"]
        assert len(agrave.components) == 0
        assert [[(p.x, p.y) for p in c] for c in agrave] == [
            [(150, 0), (450, 0), (450, 500), (150, 500)],
            [(340, 550), (300, 750), (230, 750), (300, 550)],
        ]

    def test_component_with_intermediate_master(self, ufos_and_glyphSets):
        ufos, glyphSets = ufos_and_glyphSets
        regular_glyphs, medium_glyphs, bold_glyphs = glyphSets
        assert {"dotlessi", "gravecomb", "igrave"}.issubset(regular_glyphs)
        assert {"dotlessi", "gravecomb", "igrave"}.issubset(bold_glyphs)
        assert "dotlessi" in medium_glyphs
        assert {"igrave", "gravecomb"}.isdisjoint(medium_glyphs)

        instantiator = Instantiator(
            {"Weight": (100, 100, 200)},
            [
                ({"Weight": 100}, regular_glyphs),
                ({"Weight": 150}, medium_glyphs),
                ({"Weight": 200}, bold_glyphs),
            ],
        )
        filter_ = DecomposeComponentsIFilter(include={"igrave"})

        modified = filter_(ufos, glyphSets, instantiator=instantiator)

        assert modified == {"igrave"}

        igrave = regular_glyphs["igrave"]
        assert len(igrave.components) == 0
        assert [[(p.x, p.y) for p in c] for c in igrave] == [
            [(100, 0), (200, 0), (200, 500), (100, 500)],
            [(180, 550), (150, 750), (100, 750), (150, 550)],
        ]

        # 'igrave' was also decomposed in the Medium layer, despite it was not
        # originally present; it was added by the filter and interpolated on-the-fly,
        # because Medium contained a different 'dotlessi' used as a component.
        igrave = medium_glyphs["igrave"]
        assert len(igrave.components) == 0
        assert [[(p.x, p.y) for p in c] for c in igrave] == [
            [(100, 0), (250, 0), (175, 500), (175, 500)],
            [(210, 550), (175, 750), (115, 750), (175, 550)],
        ]
        assert {"dotlessi", "igrave"}.issubset(medium_glyphs)
        assert "gravecomb" not in medium_glyphs

        igrave = bold_glyphs["igrave"]
        assert len(igrave.components) == 0
        assert [[(p.x, p.y) for p in c] for c in igrave] == [
            [(100, 0), (300, 0), (300, 500), (100, 500)],
            [(240, 550), (200, 750), (130, 750), (200, 550)],
        ]

    def test_without_instantiator(self, ufos_and_glyphSets):
        # without an instantiator (i.e. when the filter is run from the legacy
        # `compileInterpolatableTTFs` without a designspace as input but only a buch
        # of UFOs), the filter will raise a MissingComponentError while
        # trying to decompose 'agrave', because it can't interpolate the missing
        # components 'a' and 'gravecomb'
        ufos, glyphSets = ufos_and_glyphSets
        medium_glyphs = glyphSets[1]
        assert {"agrave", "dotlessi"}.issubset(medium_glyphs)
        assert {"a", "gravecomb", "igrave"}.isdisjoint(medium_glyphs)

        with pytest.raises(MissingComponentError, match="'a'"):
            DecomposeComponentsIFilter(include={"agrave"})(ufos, glyphSets)

        # the filter will not fail to decompose 'igrave' in Regular or Bold, however the
        # Medium master will not contain decomposed outlines for 'igrave', and
        # in the VF produced from these masters the 'igrave' will appear different
        # at runtime from 'dotlessi' when the Medium instance is selected.
        modified = DecomposeComponentsIFilter(include={"igrave"})(ufos, glyphSets)
        assert modified == {"igrave"}
        assert "igrave" not in medium_glyphs

    def test_locations_from_component_glyphs_get_cached(
        self, caplog, ufos_and_glyphSets
    ):
        ufos, glyphSets = ufos_and_glyphSets
        regular_glyphs, medium_glyphs, bold_glyphs = glyphSets
        instantiator = Instantiator(
            {"Weight": (100, 100, 200)},
            [
                ({"Weight": 100}, regular_glyphs),
                ({"Weight": 150}, medium_glyphs),
                ({"Weight": 200}, bold_glyphs),
            ],
        )
        philter = DecomposeComponentsIFilter()
        philter.set_context(ufos, glyphSets, instantiator)

        igrave_locations = philter.glyphSourceLocations("igrave")

        # igrave is defined only at Weight 100 and 200
        assert igrave_locations == {
            frozenset({("Weight", 100)}),
            frozenset({("Weight", 200)}),
        }

        # locationsFromComponentGlyphs logs DEBUG messages while traversing
        # recursively each component glyph
        with caplog.at_level(logging.DEBUG, logger="ufo2ft.filters"):
            component_locations = philter.locationsFromComponentGlyphs("igrave")

        assert "igrave" in caplog.text
        assert "dotlessi" in caplog.text
        assert "gravecomb" in caplog.text

        # one of igrave's components (dotlessi) is also defined at Weight 150
        expected_component_locations = igrave_locations | {frozenset({("Weight", 150)})}
        assert component_locations == expected_component_locations

        # locationsFromComponentGlyphs uses a cache to avoid traversing again component
        # glyphs that were visited before; its result isn't expected to change within
        # the current filter call.
        caplog.clear()
        with caplog.at_level(logging.DEBUG, logger="ufo2ft.filters"):
            component_locations = philter.locationsFromComponentGlyphs("igrave")

        assert "igrave" in caplog.text
        assert "dotlessi" not in caplog.text
        assert "gravecomb" not in caplog.text
        assert component_locations == expected_component_locations



import enum
import logging
import re
from io import BytesIO

from fontTools.cffLib.CFFToCFF2 import convertCFFToCFF2
from fontTools.ttLib import TTFont
from fontTools.ttLib.standardGlyphOrder import standardGlyphOrder

from ufo2ft.constants import (
    GLYPHS_DONT_USE_PRODUCTION_NAMES,
    KEEP_GLYPH_NAMES,
    USE_PRODUCTION_NAMES,
    CFFOptimization,
)

logger = logging.getLogger(__name__)


class CFFVersion(enum.IntEnum):
    CFF = 1
    CFF2 = 2


class PostProcessor:
    """Does some post-processing operations on a compiled OpenType font, using
    info from the source UFO where necessary.
    """

    GLYPH_NAME_INVALID_CHARS = re.compile("[^0-9a-zA-Z_.]")
    MAX_GLYPH_NAME_LENGTH = 63

    class SubroutinizerBackend(enum.Enum):
        COMPREFFOR = "compreffor"
        CFFSUBR = "cffsubr"

    # can override by passing explicit subroutinizer parameter to process method
    DEFAULT_SUBROUTINIZER_FOR_CFF_VERSION = {
        1: SubroutinizerBackend.CFFSUBR,
        2: SubroutinizerBackend.CFFSUBR,
    }

    def __init__(self, otf, ufo, glyphSet=None, info=None):
        self.ufo = ufo
        self.glyphSet = glyphSet if glyphSet is not None else ufo
        self.info = info
        self.otf = otf

        self._postscriptNames = ufo.lib.get("public.postscriptNames")

    def process(
        self,
        useProductionNames=None,
        optimizeCFF=True,
        cffVersion=None,
        subroutinizer=None,
    ):
        """
        useProductionNames (Optional[bool]):
          By default, when value is None, this will rename glyphs using the
          'public.postscriptNames' in then UFO lib. If the mapping is not
          present, no glyph names are renamed.

          When useProductionNames is None and the UFO lib contains the plist bool key
          "com.github.googlei18n.ufo2ft.keepGlyphNames" set to False, then the 'post'
          table is set to format 3.0 and glyph names are dropped altogether from the
          font, saving a few KBs. Note that this only works for TTF and CFF2 flavored
          fonts. We currently do not support dropping glyph names from CFF 1.0 fonts.
          When the keepGlyphNames lib key is missing or set to True, the glyph names
          will be stored in 'post' table format 2.0 for TTF and CFF2 fonts, or in the
          CFF charset.

          If useProductionNames is False, no glyphs are renamed whether or not the
          'public.postscriptNames' mapping is present.
          If the value is True, but no 'public.postscriptNames' are present,
          then uniXXXX names are generated from the glyphs' unicode.

          The 'com.github.googlei18n.ufo2ft.useProductionNames' key can be set
          in the UFO lib to control this parameter (plist boolean value).

          For legacy reasons, an alias key (with an inverted meaning) is also
          supported: "com.schriftgestaltung.Don't use Production Names";
          when this is present if the UFO lib and is set to True, this is
          equivalent to 'useProductionNames' set to False.

        optimizeCFF (bool | CFFOptimization):
          If True or >= CFFOptimization.SUBROUTINIZE, subroubtinize CFF or CFF2 table
          (if present).

        cffVersion (Optiona[int]):
          The output CFF format, choose between 1 or 2. By default, it's the same as
          as the input OTF's CFF or CFF2 table, if any. Ignored for TTFs.

        subroutinizer (Optional[str]):
          The name of the library to use for compressing CFF charstrings, if optimizeCFF
          is True and CFF or CFF2 table is present. Choose between "cffsubr" or
          "compreffor". By default "cffsubr" is used for both CFF 1 and CFF 2.
          NOTE: compreffor currently doesn't support input fonts with CFF2 table.
        """
        if self._get_cff_version(self.otf):
            if not isinstance(optimizeCFF, bool):
                optimizeCFF = optimizeCFF >= CFFOptimization.SUBROUTINIZE
            self.process_cff(
                optimizeCFF=optimizeCFF,
                cffVersion=cffVersion,
                subroutinizer=subroutinizer,
            )

        self.process_glyph_names(useProductionNames)

        if self.info:
            self.apply_fontinfo()

        return self.otf

    def process_cff(self, *, optimizeCFF=True, cffVersion=None, subroutinizer=None):
        cffInputVersion = self._get_cff_version(self.otf)
        if not cffInputVersion:
            raise ValueError("Missing required 'CFF ' or 'CFF2' table")

        if cffVersion is None:
            cffOutputVersion = cffInputVersion
        else:
            cffOutputVersion = CFFVersion(cffVersion)

        if optimizeCFF:
            if subroutinizer is None:
                backend = self.DEFAULT_SUBROUTINIZER_FOR_CFF_VERSION[cffOutputVersion]
            else:
                backend = self.SubroutinizerBackend(subroutinizer)
            self._subroutinize(backend, self.otf, cffOutputVersion)

        elif cffInputVersion != cffOutputVersion:
            if (
                cffInputVersion == CFFVersion.CFF
                and cffOutputVersion == CFFVersion.CFF2
            ):
                logger.info("Converting CFF table to CFF2")
                convertCFFToCFF2(self.otf)
            else:
                raise NotImplementedError(
                    "Unsupported CFF conversion {cffInputVersion} => {cffOutputVersion}"
                )

    def process_glyph_names(self, useProductionNames=None):
        if useProductionNames is None:
            keepGlyphNames = self.ufo.lib.get(KEEP_GLYPH_NAMES, True)
            useProductionNames = self.ufo.lib.get(
                USE_PRODUCTION_NAMES,
                not self.ufo.lib.get(GLYPHS_DONT_USE_PRODUCTION_NAMES)
                and self._postscriptNames is not None,
            )
        else:
            keepGlyphNames = True

        if keepGlyphNames:
            if "CFF " not in self.otf:
                self.set_post_table_format(self.otf, 2.0)

            if useProductionNames:
                logger.info("Renaming glyphs to final production names")
                # We need to reload the font *before* renaming glyphs, since various
                # tables may have been build/loaded using the original glyph names.
                # After reloading, we can immediately set a new glyph order and update
                # the tables (post or CFF) that stores the new postcript names; any
                # other tables that get loaded subsequently will use the new glyph names.
                self.otf = _reloadFont(self.otf)
                self._rename_glyphs_from_ufo()

        else:
            if "CFF " in self.otf:
                logger.warning(
                    "Dropping glyph names from CFF 1.0 is currently unsupported"
                )
            else:
                # To drop glyph names from TTF or CFF2, we must reload the font *after*
                # setting the post format to 3.0, since other tables may still use
                # the old glyph names.
                self.set_post_table_format(self.otf, 3.0)
                self.otf = _reloadFont(self.otf)

    def _rename_glyphs_from_ufo(self):
        """Rename glyphs using ufo.lib.public.postscriptNames in UFO."""
        rename_map = self._build_production_names()
        self.rename_glyphs(self.otf, rename_map)

    @staticmethod
    def rename_glyphs(otf, rename_map):
        newGlyphOrder = [rename_map.get(n, n) for n in otf.getGlyphOrder()]
        otf.setGlyphOrder(newGlyphOrder)

        if "post" in otf and otf["post"].formatType == 2.0:
            # annoyingly we need to update extraNames to match the new glyph order,
            # otherwise, if dumping the font to TTX directly before compiling first,
            # the post table will not contain the extraNames...
            otf["post"].extraNames = [
                g for g in newGlyphOrder if g not in standardGlyphOrder
            ]
            otf["post"].mapping = {}

        cff_tag = "CFF " if "CFF " in otf else "CFF2" if "CFF2" in otf else None
        if cff_tag == "CFF " or (cff_tag == "CFF2" and otf.isLoaded(cff_tag)):
            cff = otf[cff_tag].cff.topDictIndex[0]
            char_strings = cff.CharStrings.charStrings
            cff.CharStrings.charStrings = {
                rename_map.get(n, n): v for n, v in char_strings.items()
            }
            cff.charset = [rename_map.get(n, n) for n in cff.charset]

    def _build_production_names(self):
        seen = {}
        rename_map = {}
        for name in self.otf.getGlyphOrder():
            # Ignore glyphs that aren't in the source, as they are usually generated
            # and we lack information about them.
            if name not in self.glyphSet:
                continue
            prod_name = self._build_production_name(self.glyphSet[name])

            # strip invalid characters not allowed in postscript glyph names
            if name != prod_name:
                valid_name = self.GLYPH_NAME_INVALID_CHARS.sub("", prod_name)
                if len(valid_name) > self.MAX_GLYPH_NAME_LENGTH:
                    # if the length of the generated production name is too
                    # long, try to fall back to the original name
                    valid_name = self.GLYPH_NAME_INVALID_CHARS.sub("", name)
            else:
                valid_name = self.GLYPH_NAME_INVALID_CHARS.sub("", name)

            if len(valid_name) > self.MAX_GLYPH_NAME_LENGTH:
                logger.warning(
                    "glyph name length exceeds 63 characters: '%s'", valid_name
                )
            # add a suffix to make the production names unique
            rename_map[name] = self._unique_name(valid_name, seen)
        return rename_map

    @staticmethod
    def _unique_name(name, seen):
        """Append incremental '.N' suffix if glyph is a duplicate."""
        if name in seen:
            n = seen[name]
            while (name + ".%d" % n) in seen:
                n += 1
            seen[name] = n + 1
            name += ".%d" % n
        seen[name] = 1
        return name

    def _build_production_name(self, glyph):
        """Build a production name for a single glyph."""

        # use PostScript names from UFO lib if available
        if self._postscriptNames:
            production_name = self._postscriptNames.get(glyph.name)
            return production_name if production_name else glyph.name

        # use name derived from unicode value
        unicode_val = glyph.unicode
        if glyph.unicode is not None:
            return "{}{:04X}".format(
                "u" if unicode_val > 0xFFFF else "uni", unicode_val
            )

        # use production name + last (non-script) suffix if possible
        parts = glyph.name.rsplit(".", 1)
        if len(parts) == 2 and parts[0] in self.glyphSet:
            return "{}.{}".format(
                self._build_production_name(self.glyphSet[parts[0]]),
                parts[1],
            )

        # use ligature name, making sure to look up components with suffixes
        parts = glyph.name.split(".", 1)
        if len(parts) == 2:
            liga_parts = ["{}.{}".format(n, parts[1]) for n in parts[0].split("_")]
        else:
            liga_parts = glyph.name.split("_")
        if len(liga_parts) > 1 and all(n in self.glyphSet for n in liga_parts):
            unicode_vals = [self.glyphSet[n].unicode for n in liga_parts]
            if all(v and v <= 0xFFFF for v in unicode_vals):
                return "uni" + "".join("%04X" % v for v in unicode_vals)
            return "_".join(
                self._build_production_name(self.glyphSet[n]) for n in liga_parts
            )

        return glyph.name

    @staticmethod
    def set_post_table_format(otf, formatType):
        if formatType not in (2.0, 3.0):
            raise NotImplementedError(formatType)

        post = otf.get("post")
        if post:
            if post.formatType != formatType:
                logger.info("Setting post.formatType = %s", formatType)
                post.formatType = formatType
            # we want to update extraNames list even if formatType is the same
            # so we don't have to reload the font
            if formatType == 2.0:
                post.extraNames = [
                    g for g in otf.getGlyphOrder() if g not in standardGlyphOrder
                ]
                post.mapping = {}
            else:
                for attr in ("extraNames", "mapping"):
                    if hasattr(post, attr):
                        delattr(post, attr)
                post.glyphOrder = None

    @staticmethod
    def _get_cff_version(otf):
        if "CFF " in otf:
            return CFFVersion.CFF
        elif "CFF2" in otf:
            return CFFVersion.CFF2
        else:
            return None

    @classmethod
    def _subroutinize(cls, backend, otf, cffVersion):
        subroutinize = getattr(cls, f"_subroutinize_with_{backend.value}")
        subroutinize(otf, cffVersion)

    @classmethod
    def _subroutinize_with_compreffor(cls, otf, cffVersion):
        from compreffor import compress

        if cls._get_cff_version(otf) != CFFVersion.CFF or cffVersion != CFFVersion.CFF:
            raise NotImplementedError(
                "Only 'CFF ' 1.0 is supported by compreffor; try using cffsubr"
            )

        logger.info("Subroutinizing CFF table with compreffor")

        compress(otf)

    @classmethod
    def _subroutinize_with_cffsubr(cls, otf, cffVersion):
        import cffsubr

        cffInputVersion = cls._get_cff_version(otf)
        assert cffInputVersion is not None, "Missing required 'CFF ' or 'CFF2' table"

        msg = f"Subroutinizing {cffInputVersion.name} table with cffsubr"
        if cffInputVersion != cffVersion:
            msg += f" (output format: {cffVersion.name})"
        logger.info(msg)

        return cffsubr.subroutinize(otf, cff_version=cffVersion, keep_glyph_names=False)

    def apply_fontinfo(self):
        """Apply the fontinfo data from the DesignSpace variable-font's lib to
        the compiled font."""
        from ufo2ft.infoCompiler import InfoCompiler

        logger.info("Applying variable-font info from DesignSpace lib")

        compiler = InfoCompiler(self.otf, self.ufo, self.info)
        compiler.compile()


def _reloadFont(font: TTFont) -> TTFont:
    """Recompile a font to arrive at the final internal layout."""
    stream = BytesIO()
    font.save(stream)
    stream.seek(0)
    # keep the same Config (constructor will make a copy)
    return TTFont(stream, cfg=font.cfg)



from dataclasses import dataclass
from typing import Optional, Type

from ufo2ft.constants import CFFOptimization
from ufo2ft.outlineCompiler import OutlineOTFCompiler
from ufo2ft.preProcessor import OTFPreProcessor

from .baseCompiler import BaseCompiler


@dataclass
class OTFCompiler(BaseCompiler):
    preProcessorClass: Type = OTFPreProcessor
    outlineCompilerClass: Type = OutlineOTFCompiler
    optimizeCFF: CFFOptimization = CFFOptimization.SUBROUTINIZE
    roundTolerance: Optional[float] = None
    cffVersion: int = 1
    subroutinizer: Optional[str] = None
